"""Functions related to layer eigenmode calculation for the FMM algorithm.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import dataclasses
from typing import Tuple

import jax
import jax.numpy as jnp

from fmmax import basis, fmm, utils


@dataclasses.dataclass
class LayerSolveResult:
    """Stores the result of a layer eigensolve.

    This eigenvalue problem is specified in equation 28 of [2012 Liu].

    Attributes:
        wavelength: The wavelength for the solve.
        in_plane_wavevector: The in-plane wavevector for the solve.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        expansion: The expansion used for the eigensolve.
        eigenvalues: The layer eigenvalues.
        eigenvectors: The layer eigenvectors.
        eta_matrix: The fourier-transformed inverse of zz-component of permittivity.
        z_permittivity_matrix: The fourier-transformed zz-component of permittivity.
        omega_script_k_matrix: The omega-script-k matrix from equation 26 of
            [2012 Liu], which is needed to generate the layer scattering matrix.
    """

    wavelength: jnp.ndarray
    in_plane_wavevector: jnp.ndarray
    primitive_lattice_vectors: basis.LatticeVectors
    expansion: basis.Expansion
    eigenvalues: jnp.ndarray
    eigenvectors: jnp.ndarray
    eta_matrix: jnp.ndarray
    z_permittivity_matrix: jnp.ndarray
    omega_script_k_matrix: jnp.ndarray

    @property
    def batch_shape(self) -> Tuple[int, ...]:
        return self.eigenvectors.shape[:-2]

    def __post_init__(self) -> None:
        """Validates shapes of the `LayerSolveResult` attributes."""
        if self.wavelength.ndim != len(
            self.batch_shape
        ) or not utils.batch_compatible_shapes(self.wavelength.shape, self.batch_shape):
            raise ValueError(
                f"`wavelength` must have compatible batch shape, but got shape {self.wavelength.shape} "
                f"when `eigenvectors` shape is {self.eigenvectors.shape}."
            )
        if self.in_plane_wavevector.ndim != len(
            self.batch_shape
        ) + 1 or not utils.batch_compatible_shapes(
            self.in_plane_wavevector.shape[:-1], self.batch_shape
        ):
            raise ValueError(
                f"`in_plane_wavevector` must have compatible batch shape, but got shape "
                f"{self.in_plane_wavevector.shape} when `eigenvectors` shape is {self.eigenvectors.shape}."
            )
        if self.expansion.num_terms * 2 != self.eigenvectors.shape[-1]:
            raise ValueError(
                f"`eigenvectors` must have shape compatible with `expansion.num_terms`, but got shape "
                f"{self.eigenvectors.shape} when `num_terms` shape is {self.expansion.num_terms}."
            )
        if self.eigenvalues.shape != self.eigenvectors.shape[:-1]:
            raise ValueError(
                f"`eigenvalues` must have compatible shape, but got shape {self.eigenvalues.shape} "
                f"when `eigenvectors` shape is {self.eigenvectors.shape}."
            )

        expected_matrix_shape = self.batch_shape + (self.expansion.num_terms,) * 2
        if self.eta_matrix.ndim != len(
            expected_matrix_shape
        ) or not utils.batch_compatible_shapes(
            self.eta_matrix.shape, expected_matrix_shape
        ):
            raise ValueError(
                f"`eta_matrix` must have shape compatible with `eigenvectors`, but got "
                f"shapes {self.eta_matrix.shape}  and {self.eigenvectors.shape}."
            )
        if self.z_permittivity_matrix.ndim != len(
            expected_matrix_shape
        ) or not utils.batch_compatible_shapes(
            self.z_permittivity_matrix.shape, expected_matrix_shape
        ):
            raise ValueError(
                f"`z_permittivity_matrix` must have shape compatible with `eigenvectors`, but got "
                f"shapes {self.z_permittivity_matrix.shape}  and {self.eigenvectors.shape}."
            )
        if self.omega_script_k_matrix.shape != self.eigenvectors.shape:
            raise ValueError(
                f"`omega_script_k_matrix` must have shape matching `eigenvectors`, but got "
                f"shapes {self.omega_script_k_matrix.shape}  and {self.eigenvectors.shape}."
            )


def eigensolve_isotropic_media(
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivity: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: fmm.Formulation,
) -> LayerSolveResult:
    """Performs the eigensolve for a layer with isotropic permittivity.

    This function performs either a uniform-layer or patterned-layer eigensolve,
    depending on the shape of the trailing dimensions of a given layer permittivity.
    When the final two dimensions have shape `(1, 1)`, the layer is treated as
    uniform. Otherwise, it is patterned.

    Args:
        wavelength: The free space wavelength of the excitation.
        in_plane_wavevector: `(kx0, ky0)`.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivity: The permittivity array.
        expansion: The field expansion to be used.
        formulation: Specifies the formulation to be used.

    Returns:
        The `LayerSolveResult`.
    """
    if permittivity.shape[-2:] == (1, 1):
        return eigensolve_uniform_isotropic_media(
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            permittivity=permittivity,
            expansion=expansion,
        )
    else:
        return eigensolve_patterned_isotropic_media(
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            permittivity=permittivity,
            expansion=expansion,
            formulation=formulation,
        )


def eigensolve_anisotropic_media(
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivity_xx: jnp.ndarray,
    permittivity_xy: jnp.ndarray,
    permittivity_yx: jnp.ndarray,
    permittivity_yy: jnp.ndarray,
    permittivity_zz: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: fmm.Formulation,
) -> LayerSolveResult:
    """Performs the eigensolve for a layer with anisotropic permittivity.

    This function performs either a uniform-layer or patterned-layer eigensolve,
    depending on the shape of the trailing dimensions of a given layer permittivity.
    When the final two dimensions have shape `(1, 1)`, the layer is treated as
    uniform. Otherwise, it is patterned.

    Args:
        wavelength: The free space wavelength of the excitation.
        in_plane_wavevector: `(kx0, ky0)`.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivity_xx: The xx-component of the permittivity tensor, with
            shape `(..., nx, ny)`.
        permittivity_xy: The xy-component of the permittivity tensor.
        permittivity_yx: The yx-component of the permittivity tensor.
        permittivity_yy: The yy-component of the permittivity tensor.
        permittivity_zz: The zz-component of the permittivity tensor.
        expansion: The field expansion to be used.
        formulation: Specifies the formulation to be used.

    Returns:
        The `LayerSolveResult`.
    """
    if permittivity_xx.shape[-2:] == (1, 1):
        return eigensolve_uniform_anisotropic_media(
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            permittivity_xx=permittivity_xx,
            permittivity_xy=permittivity_xy,
            permittivity_yx=permittivity_yx,
            permittivity_yy=permittivity_yy,
            permittivity_zz=permittivity_zz,
            expansion=expansion,
        )
    else:
        return eigensolve_patterned_anisotropic_media(
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            permittivity_xx=permittivity_xx,
            permittivity_xy=permittivity_xy,
            permittivity_yx=permittivity_yx,
            permittivity_yy=permittivity_yy,
            permittivity_zz=permittivity_zz,
            expansion=expansion,
            formulation=formulation,
        )


def eigensolve_uniform_isotropic_media(
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivity: jnp.ndarray,
    expansion: basis.Expansion,
) -> LayerSolveResult:
    r"""Returns the the results of a uniform isotropic layer eigensolve.

    The layer is uniform and isotropic, in the sense that the permittivity does not
    vary spatially and has no orientation dependence.

    Args:
        wavelength: The free space wavelength of the excitation.
        in_plane_wavevector: `(kx0, ky0)`.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivity: The scalar permittivity for the layer, with shape `(..., 1, 1)`.
        expansion: The field expansion to be used.

    Returns:
        The `LayerSolveResult`.
    """
    wavelength, in_plane_wavevector, permittivity = _validate_and_broadcast(
        wavelength, in_plane_wavevector, permittivity
    )
    if permittivity.shape[-2:] != (1, 1):
        raise ValueError(
            f"Trailing axes of `permittivity` must have shape (1, 1) but got a shape "
            f"of {permittivity.shape}."
        )

    batch_shape = jnp.broadcast_shapes(
        wavelength.shape, in_plane_wavevector.shape[:-1], permittivity.shape[:-2]
    )

    num_eigenvalues = 2 * expansion.num_terms
    permittivity = jnp.squeeze(permittivity, axis=(-2, -1))

    # Transverse wavevectors are the `k + G` from equation 5 of [2012 Liu].
    transverse_wavevectors = basis.transverse_wavevectors(
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
    )

    # In uniform media, the eigenvectors are just the plane waves.
    eigenvectors = jnp.broadcast_to(
        jnp.eye(num_eigenvalues, dtype=complex),
        shape=batch_shape + (num_eigenvalues, num_eigenvalues),
    )

    angular_frequency = utils.angular_frequency_for_wavelength(wavelength)
    kx = transverse_wavevectors[..., 0]
    ky = transverse_wavevectors[..., 1]
    eigenvalues = jnp.sqrt(
        (
            permittivity[..., jnp.newaxis] * angular_frequency[..., jnp.newaxis] ** 2
            - kx**2
            - ky**2
        ).astype(complex)
    )
    eigenvalues = _select_eigenvalues_sign(eigenvalues)
    eigenvalues = jnp.tile(eigenvalues, 2)

    batch_shape = eigenvalues.shape[:-1]
    eta_matrix = jnp.broadcast_to(
        1 / permittivity[..., jnp.newaxis], batch_shape + (expansion.num_terms,)
    )
    eta_matrix = utils.diag(eta_matrix)
    z_permittivity_matrix = jnp.broadcast_to(
        permittivity[..., jnp.newaxis], batch_shape + (expansion.num_terms,)
    )
    z_permittivity_matrix = utils.diag(z_permittivity_matrix)

    # The matrix from equation 26 of [2012 Liu].
    angular_frequency_squared = angular_frequency[..., jnp.newaxis, jnp.newaxis] ** 2
    angular_frequency_squared *= jnp.eye(num_eigenvalues)
    omega_script_k_matrix = angular_frequency_squared - _script_k_matrix_uniform(
        permittivity, transverse_wavevectors
    )
    return LayerSolveResult(
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
        eigenvalues=eigenvalues,
        eigenvectors=eigenvectors,
        eta_matrix=eta_matrix,
        z_permittivity_matrix=z_permittivity_matrix,
        omega_script_k_matrix=omega_script_k_matrix,
    )


def eigensolve_uniform_anisotropic_media(
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivity_xx: jnp.ndarray,
    permittivity_xy: jnp.ndarray,
    permittivity_yx: jnp.ndarray,
    permittivity_yy: jnp.ndarray,
    permittivity_zz: jnp.ndarray,
    expansion: basis.Expansion,
) -> LayerSolveResult:
    """Returns the results of a uniform anisotropic layer eigensolve.

    The layer is uniform and anisotropic, in the sense that the permittivity does not
    vary spatially and has orientation dependence.

    Args:
        wavelength: The free space wavelength of the excitation.
        in_plane_wavevector: `(kx0, ky0)`.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivity_xx: The xx-component of the permittivity tensor, with
            shape `(..., nx, ny)`.
        permittivity_xy: The xy-component of the permittivity tensor.
        permittivity_yx: The yx-component of the permittivity tensor.
        permittivity_yy: The yy-component of the permittivity tensor.
        permittivity_zz: The zz-component of the permittivity tensor.
        expansion: The field expansion to be used.

    Returns:
        The `LayerSolveResult`.
    """
    (
        wavelength,
        in_plane_wavevector,
        permittivity_xx,
        permittivity_xy,
        permittivity_yx,
        permittivity_yy,
        permittivity_zz,
    ) = _validate_and_broadcast(
        wavelength,
        in_plane_wavevector,
        permittivity_xx,
        permittivity_xy,
        permittivity_yx,
        permittivity_yy,
        permittivity_zz,
    )
    if permittivity_xx.shape[-2:] != (1, 1):
        raise ValueError(
            f"Trailing axes of `permittivity_xx` must have shape (1, 1) but got a shape "
            f"of {permittivity_xx.shape}."
        )

    batch_shape = jnp.broadcast_shapes(
        wavelength.shape, in_plane_wavevector.shape[:-1], permittivity_xx.shape[:-2]
    )
    shape = batch_shape + (expansion.num_terms,)
    permittivity_xx = jnp.broadcast_to(jnp.squeeze(permittivity_xx, axis=-1), shape)
    permittivity_xy = jnp.broadcast_to(jnp.squeeze(permittivity_xy, axis=-1), shape)
    permittivity_yx = jnp.broadcast_to(jnp.squeeze(permittivity_yx, axis=-1), shape)
    permittivity_yy = jnp.broadcast_to(jnp.squeeze(permittivity_yy, axis=-1), shape)
    permittivity_zz = jnp.broadcast_to(jnp.squeeze(permittivity_zz, axis=-1), shape)

    eta_matrix = utils.diag(1 / permittivity_zz)
    z_permittivity_matrix = utils.diag(permittivity_zz)
    transverse_permittivity_matrix = jnp.block(
        [
            [utils.diag(permittivity_xx), utils.diag(permittivity_xy)],
            [utils.diag(permittivity_yx), utils.diag(permittivity_yy)],
        ]
    )

    return _eigensolve_patterned_media(
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        eta_matrix=eta_matrix,
        z_permittivity_matrix=z_permittivity_matrix,
        transverse_permittivity_matrix=transverse_permittivity_matrix,
        expansion=expansion,
    )


# -----------------------------------------------------------------------------
# Patterned layer eigensolves.
# -----------------------------------------------------------------------------


def eigensolve_patterned_isotropic_media(
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivity: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: fmm.Formulation,
) -> LayerSolveResult:
    r"""Returns the results of a patterned isotropic layer eigensolve.

    The layer is patterned and isotropic, in the sense that the permittivity varies
    spatially and has no orientation dependence.

    Args:
        wavelength: The free space wavelength of the excitation.
        in_plane_wavevector: `(kx0, ky0)`.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivity: The permittivity array.
        expansion: The field expansion to be used.
        formulation: Specifies the formulation to be used.

    Returns:
        The `LayerSolveResult`.
    """
    wavelength, in_plane_wavevector, permittivity = _validate_and_broadcast(
        wavelength, in_plane_wavevector, permittivity
    )
    (
        eta_matrix,
        z_permittivity_matrix,
        transverse_permittivity_matrix,
    ) = fmm.fourier_matrices_patterned_isotropic_media(
        primitive_lattice_vectors=primitive_lattice_vectors,
        permittivity=permittivity,
        expansion=expansion,
        formulation=formulation,
    )
    return _eigensolve_patterned_media(
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        eta_matrix=eta_matrix,
        z_permittivity_matrix=z_permittivity_matrix,
        transverse_permittivity_matrix=transverse_permittivity_matrix,
        expansion=expansion,
    )


def eigensolve_patterned_anisotropic_media(
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivity_xx: jnp.ndarray,
    permittivity_xy: jnp.ndarray,
    permittivity_yx: jnp.ndarray,
    permittivity_yy: jnp.ndarray,
    permittivity_zz: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: fmm.Formulation,
) -> LayerSolveResult:
    """Returns the results of a patterned anisotropic layer eigensolve.

    The layer is patterned and anisotropic, in the sense that the permittivity varies
    spatially and has orientation dependence.

    Args:
        wavelength: The free space wavelength of the excitation.
        in_plane_wavevector: `(kx0, ky0)`.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivity_xx: The xx-component of the permittivity tensor, with
            shape `(..., nx, ny)`.
        permittivity_xy: The xy-component of the permittivity tensor.
        permittivity_yx: The yx-component of the permittivity tensor.
        permittivity_yy: The yy-component of the permittivity tensor.
        permittivity_zz: The zz-component of the permittivity tensor.
        expansion: The field expansion to be used.
        formulation: Specifies the formulation to be used.

    Returns:
        The `LayerSolveResult`.
    """
    (
        wavelength,
        in_plane_wavevector,
        permittivity_xx,
        permittivity_xy,
        permittivity_yx,
        permittivity_yy,
        permittivity_zz,
    ) = _validate_and_broadcast(
        wavelength,
        in_plane_wavevector,
        permittivity_xx,
        permittivity_xy,
        permittivity_yx,
        permittivity_yy,
        permittivity_zz,
    )
    (
        eta_matrix,
        z_permittivity_matrix,
        transverse_permittivity_matrix,
    ) = fmm.fourier_matrices_patterned_anisotropic_media(
        primitive_lattice_vectors=primitive_lattice_vectors,
        permittivity_xx=permittivity_xx,
        permittivity_xy=permittivity_xy,
        permittivity_yx=permittivity_yx,
        permittivity_yy=permittivity_yy,
        permittivity_zz=permittivity_zz,
        expansion=expansion,
        formulation=formulation,
    )
    return _eigensolve_patterned_media(
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        eta_matrix=eta_matrix,
        z_permittivity_matrix=z_permittivity_matrix,
        transverse_permittivity_matrix=transverse_permittivity_matrix,
        expansion=expansion,
    )


# -----------------------------------------------------------------------------
# Helper function for patterned layer eigensolve.
# -----------------------------------------------------------------------------


def _eigensolve_patterned_media(
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    eta_matrix: jnp.ndarray,
    z_permittivity_matrix: jnp.ndarray,
    transverse_permittivity_matrix: jnp.ndarray,
    expansion: basis.Expansion,
) -> LayerSolveResult:
    r"""Returns the results of a patterned isotropic layer eigensolve.

    The layer is patterned and isotropic, in the sense that the permittivity varies
    spatially and has no orientation dependence.

    Args:
        wavelength: The free space wavelength of the excitation.
        in_plane_wavevector: `(kx0, ky0)`.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        eta_matrix: The fourier-transformed inverse of zz-component of permittivity.
        z_permittivity_matrix: The fourier-transformed zz-component of permittivity.
        transverse_permittivity_matrix: The fourier-transformed transverse permittivity
            matrix from equation 15 of [2012 Liu].
        expansion: The field expansion to be used.

    Returns:
        The `LayerSolveResult`.
    """
    num_eigenvalues = 2 * expansion.basis_coefficients.shape[0]

    # Transverse wavevectors are the `k + G` from equation 5 of [2012 Liu].
    transverse_wavevectors = basis.transverse_wavevectors(
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
    )

    # The matrix from equation 11 of [2012 Liu].
    angular_frequency = utils.angular_frequency_for_wavelength(wavelength)

    # The k-matrix from equation 23 of [2012 Liu].
    k_matrix = _k_matrix(transverse_wavevectors)

    # The matrix from equation 26 of [2012 Liu].
    angular_frequency_squared = angular_frequency[..., jnp.newaxis, jnp.newaxis] ** 2
    angular_frequency_squared *= jnp.eye(num_eigenvalues)
    omega_script_k_matrix = angular_frequency_squared - _script_k_matrix_patterned(
        z_permittivity_matrix, transverse_wavevectors
    )
    matrix = transverse_permittivity_matrix @ omega_script_k_matrix - k_matrix
    eigenvalues_squared, eigenvectors = utils.eig(matrix)
    eigenvalues = jnp.sqrt(eigenvalues_squared)
    eigenvalues = _select_eigenvalues_sign(eigenvalues)
    return LayerSolveResult(
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
        eigenvalues=eigenvalues,
        eigenvectors=eigenvectors,
        eta_matrix=eta_matrix,
        z_permittivity_matrix=z_permittivity_matrix,
        omega_script_k_matrix=omega_script_k_matrix,
    )


# -----------------------------------------------------------------------------
# Helper functions for validation and matrix assembly.
# -----------------------------------------------------------------------------


def _validate_and_broadcast(
    angular_frequency: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    *permittivities: jnp.ndarray,
) -> Tuple[jnp.ndarray, ...]:
    """Validates that shapes are compatible and adds required batch dimensions."""
    if not in_plane_wavevector.shape[-1] == 2:
        raise ValueError(
            f"`in_plane_wavevector` must have a final dimension of size 2 but got "
            f"a shape of {in_plane_wavevector.shape}."
        )

    if not all([permittivities[0].shape == p.shape for p in permittivities]):
        raise ValueError("Got permittivities with differing shapes.")

    permittivity = permittivities[0]
    if not utils.batch_compatible_shapes(
        angular_frequency.shape,
        in_plane_wavevector.shape[:-1],
        permittivity.shape[:-2],
    ):
        raise ValueError(
            f"`angular_frequency`, `in_plane_wavevector`, and `permittivity` "
            f"must be batch-compatible, but got shapes of {angular_frequency.shape}, "
            f"{in_plane_wavevector.shape}, and {permittivity.shape}."
        )

    num_batch_dims = max(
        [
            angular_frequency.ndim,
            in_plane_wavevector.ndim - 1,
            permittivity.ndim - 2,
        ]
    )
    angular_frequency = utils.atleast_nd(angular_frequency, n=num_batch_dims)
    in_plane_wavevector = utils.atleast_nd(in_plane_wavevector, n=num_batch_dims + 1)

    permittivities = tuple(
        [utils.atleast_nd(p, n=num_batch_dims + 2) for p in permittivities]
    )
    return (angular_frequency, in_plane_wavevector) + permittivities


def _select_eigenvalues_sign(eigenvalues: jnp.ndarray) -> jnp.ndarray:
    """Selects the sign of eigenvalues to have strictly positive imaginary part.

    Args:
        eigenvalues: The eigenvalues whose sign is to be adjusted.

    Returns:
        The eigenvalues with adjusted sign.
    """
    return jnp.where(jnp.imag(eigenvalues) < 0, -eigenvalues, eigenvalues)


def _script_k_matrix_patterned(
    z_permittivity_matrix: jnp.ndarray,
    transverse_wavevectors: jnp.ndarray,
) -> jnp.ndarray:
    """Returns the patterned-layer script-k matrix from eq. 19 of [2012 Liu]."""
    kx = transverse_wavevectors[..., 0]
    ky = transverse_wavevectors[..., 1]
    z_inv_kx = jnp.linalg.solve(z_permittivity_matrix, utils.diag(kx))
    z_inv_ky = jnp.linalg.solve(z_permittivity_matrix, utils.diag(ky))
    return jnp.block(
        [
            [ky[..., :, jnp.newaxis] * z_inv_ky, -ky[..., :, jnp.newaxis] * z_inv_kx],
            [-kx[..., :, jnp.newaxis] * z_inv_ky, kx[..., :, jnp.newaxis] * z_inv_kx],
        ]
    )


def _script_k_matrix_uniform(
    permittivity: jnp.ndarray,
    transverse_wavevectors: jnp.ndarray,
) -> jnp.ndarray:
    """Returns the uniform-layer script-k matrix from eq. 19 of [2012 Liu]."""
    kx = transverse_wavevectors[..., 0]
    ky = transverse_wavevectors[..., 1]
    return jnp.block(
        [
            [
                utils.diag(ky / permittivity[..., jnp.newaxis] * ky),
                utils.diag(-ky / permittivity[..., jnp.newaxis] * kx),
            ],
            [
                utils.diag(-kx / permittivity[..., jnp.newaxis] * ky),
                utils.diag(kx / permittivity[..., jnp.newaxis] * kx),
            ],
        ]
    )


def _k_matrix(transverse_wavevectors: jnp.ndarray) -> jnp.ndarray:
    """Returns the k matrix from equation 23 of [2012 Liu]."""
    kx = transverse_wavevectors[..., 0]
    ky = transverse_wavevectors[..., 1]
    return jnp.block(
        [
            [utils.diag(kx**2), utils.diag(kx * ky)],
            [utils.diag(ky * kx), utils.diag(ky**2)],
        ]
    )


# -----------------------------------------------------------------------------
# Register custom objects in this module with jax to enable `jit`.
# -----------------------------------------------------------------------------


jax.tree_util.register_pytree_node(
    LayerSolveResult,
    lambda x: (
        (
            x.wavelength,
            x.in_plane_wavevector,
            x.primitive_lattice_vectors,
            x.expansion,
            x.eigenvalues,
            x.eigenvectors,
            x.eta_matrix,
            x.z_permittivity_matrix,
            x.omega_script_k_matrix,
        ),
        None,
    ),
    lambda _, x: LayerSolveResult(*x),
)



"""Tests for `fmmax.scattering`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import dataclasses
import unittest

import grcwa
import jax
import jax.numpy as jnp
import numpy as onp

from fmmax import basis, fmm, layer, scattering, sources

# Enable 64-bit precision for higher accuracy.
jax.config.update("jax_enable_x64", True)


WAVELENGTH = jnp.asarray(0.628)
PRIMITIVE_LATTICE_VECTORS = basis.LatticeVectors(
    u=onp.asarray((0.9, 0.1)), v=onp.asarray((0.2, 0.8))
)
EXPANSION = basis.generate_expansion(
    primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
    approximate_num_terms=4,
    truncation=basis.Truncation.CIRCULAR,
)
IN_PLANE_WAVEVECTOR = basis.plane_wave_in_plane_wavevector(
    wavelength=WAVELENGTH,
    polar_angle=jnp.pi * 0.1,
    azimuthal_angle=jnp.pi * 0.2,
    permittivity=jnp.asarray(1.0),
)


def _random_normal_complex(key, shape):
    x = jax.random.normal(key, (2,) + shape)
    return x[0, ...] + 1j * x[1, ...]


def _dummy_solve_result(
    key,
    wavelength=WAVELENGTH,
    in_plane_wavevector=IN_PLANE_WAVEVECTOR,
    primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
    expansion=EXPANSION,
):
    key_eigenvalues, key_eigenvectors, key_omega_k = jax.random.split(key, 3)
    dim = expansion.basis_coefficients.shape[0]
    return layer.LayerSolveResult(
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
        eigenvalues=_random_normal_complex(key_eigenvalues, (2 * dim,)),
        eigenvectors=_random_normal_complex(key_eigenvectors, (2 * dim, 2 * dim)),
        eta_matrix=_random_normal_complex(key_eigenvectors, (dim, dim)),
        z_permittivity_matrix=_random_normal_complex(key_eigenvectors, (dim, dim)),
        omega_script_k_matrix=_random_normal_complex(key_omega_k, (2 * dim, 2 * dim)),
    )


def _stack_solve_result(
    key,
    wavelength=WAVELENGTH,
    in_plane_wavevector=IN_PLANE_WAVEVECTOR,
    primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
    expansion=EXPANSION,
):
    permittivities = [
        jnp.array([[1.0 + 0.0j]]),
        jnp.array([[1.0 + 0.0j]]),
        jax.random.normal(key, (128, 128)) * 0.1 + (3.0 + 0.0j),
        jnp.array([[5.0 + 0.0j]]),
        jnp.array([[5.0 + 0.0j]]),
    ]
    return [
        layer.eigensolve_isotropic_media(
            wavelength,
            in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            permittivity=p,
            expansion=expansion,
            formulation=fmm.Formulation.FFT,
        )
        for p in permittivities
    ]


class ScatteringMatrixTest(unittest.TestCase):
    def test_compare_to_grcwa(self):
        solve_results = [
            _dummy_solve_result(jax.random.PRNGKey(0)),
            _dummy_solve_result(jax.random.PRNGKey(1)),
            _dummy_solve_result(jax.random.PRNGKey(2)),
            _dummy_solve_result(jax.random.PRNGKey(3)),
        ]
        thicknesses = [1.0, 1.5, 2.0, 2.5]
        result = scattering.stack_s_matrix(solve_results, thicknesses)
        expected_s11, expected_s12, expected_s21, expected_s22 = grcwa.rcwa.GetSMatrix(
            indi=0,
            indj=len(solve_results) - 1,
            q_list=[r.eigenvalues for r in solve_results],
            phi_list=[r.eigenvectors for r in solve_results],
            kp_list=[r.omega_script_k_matrix for r in solve_results],
            thickness_list=thicknesses,
        )
        with self.subTest("s11"):
            onp.testing.assert_allclose(result.s11, expected_s11)
        with self.subTest("s12"):
            onp.testing.assert_allclose(result.s12, expected_s12)
        with self.subTest("s21"):
            onp.testing.assert_allclose(result.s21, expected_s21)
        with self.subTest("s22"):
            onp.testing.assert_allclose(result.s22, expected_s22)

    def test_compare_to_grcwa_actual_solve(self):
        solve_results = _stack_solve_result(jax.random.PRNGKey(0))
        thicknesses = [1.0, 1.5, 2.0, 2.5, 1.0]
        result = scattering.stack_s_matrix(solve_results, thicknesses)
        expected_s11, expected_s12, expected_s21, expected_s22 = grcwa.rcwa.GetSMatrix(
            indi=0,
            indj=len(solve_results) - 1,
            q_list=[r.eigenvalues for r in solve_results],
            phi_list=[r.eigenvectors for r in solve_results],
            kp_list=[r.omega_script_k_matrix for r in solve_results],
            thickness_list=thicknesses,
        )
        with self.subTest("s11"):
            onp.testing.assert_allclose(result.s11, expected_s11)
        with self.subTest("s12"):
            onp.testing.assert_allclose(result.s12, expected_s12)
        with self.subTest("s21"):
            onp.testing.assert_allclose(result.s21, expected_s21)
        with self.subTest("s22"):
            onp.testing.assert_allclose(result.s22, expected_s22)

    def test_construct_s_matrix_with_prepend(self):
        solve_results = [
            _dummy_solve_result(jax.random.PRNGKey(0)),
            _dummy_solve_result(jax.random.PRNGKey(1)),
            _dummy_solve_result(jax.random.PRNGKey(2)),
            _dummy_solve_result(jax.random.PRNGKey(3)),
        ]
        thicknesses = [1.0, 1.5, 2.0, 2.5]
        expected = scattering.stack_s_matrix(solve_results, thicknesses)

        s_matrix = scattering.stack_s_matrix(solve_results[-2:], thicknesses[-2:])
        s_matrix = scattering.prepend_layer(s_matrix, solve_results[1], thicknesses[1])
        s_matrix = scattering.prepend_layer(s_matrix, solve_results[0], thicknesses[0])

        with self.subTest("s11"):
            onp.testing.assert_allclose(s_matrix.s11, expected.s11)
        with self.subTest("s12"):
            onp.testing.assert_allclose(s_matrix.s12, expected.s12)
        with self.subTest("s21"):
            onp.testing.assert_allclose(s_matrix.s21, expected.s21)
        with self.subTest("s22"):
            onp.testing.assert_allclose(s_matrix.s22, expected.s22)

    def test_interior_s_matrices(self):
        solve_results = [
            _dummy_solve_result(jax.random.PRNGKey(0)),
            _dummy_solve_result(jax.random.PRNGKey(1)),
            _dummy_solve_result(jax.random.PRNGKey(2)),
            _dummy_solve_result(jax.random.PRNGKey(3)),
        ]
        thicknesses = [1.0, 1.5, 2.0, 2.5]
        s_matrices_interior = scattering.stack_s_matrices_interior(
            solve_results, thicknesses
        )

        # Test that that the "above" s-matrix for the first layer
        # and the "below" s-matrix for the last layer are equal.
        a, b = (s_matrices_interior[0][1], s_matrices_interior[-1][0])
        with self.subTest("s11-above/below"):
            onp.testing.assert_allclose(a.s11, b.s11)
        with self.subTest("s12-above/below"):
            onp.testing.assert_allclose(a.s12, b.s12)
        with self.subTest("s21-above/below"):
            onp.testing.assert_allclose(a.s21, b.s21)
        with self.subTest("s22-above/below"):
            onp.testing.assert_allclose(a.s22, b.s22)

        # Test that that the "below" s-matrix for the first layer
        # and the "above" s-matrix for the last layer are equal.
        a, b = (s_matrices_interior[0][0], s_matrices_interior[-1][1])
        with self.subTest("s11-below/above"):
            onp.testing.assert_allclose(a.s11, b.s11)
        with self.subTest("s12-below/above"):
            onp.testing.assert_allclose(a.s12, b.s12)
        with self.subTest("s21-below/above"):
            onp.testing.assert_allclose(a.s21, b.s21)
        with self.subTest("s22-below/above"):
            onp.testing.assert_allclose(a.s22, b.s22)

    def test_interior_s_matrices_actual_solve(self):
        solve_results = _stack_solve_result(jax.random.PRNGKey(0))
        thicknesses = [1.0, 1.5, 2.0, 2.5, 1.0]
        s_matrices_interior = scattering.stack_s_matrices_interior(
            solve_results, thicknesses
        )

        # Test that that the "above" s-matrix for the first layer
        # and the "below" s-matrix for the last layer are equal.
        a, b = (s_matrices_interior[0][1], s_matrices_interior[-1][0])
        with self.subTest("s11-above/below"):
            onp.testing.assert_allclose(a.s11, b.s11)
        with self.subTest("s12-above/below"):
            onp.testing.assert_allclose(a.s12, b.s12)
        with self.subTest("s21-above/below"):
            onp.testing.assert_allclose(a.s21, b.s21)
        with self.subTest("s22-above/below"):
            onp.testing.assert_allclose(a.s22, b.s22)

        # Test that that the "below" s-matrix for the first layer
        # and the "above" s-matrix for the last layer are equal.
        a, b = (s_matrices_interior[0][0], s_matrices_interior[-1][1])
        with self.subTest("s11-below/above"):
            onp.testing.assert_allclose(a.s11, b.s11)
        with self.subTest("s12-below/above"):
            onp.testing.assert_allclose(a.s12, b.s12)
        with self.subTest("s21-below/above"):
            onp.testing.assert_allclose(a.s21, b.s21)
        with self.subTest("s22-below/above"):
            onp.testing.assert_allclose(a.s22, b.s22)

    def test_can_jit(self):
        solve_results = [
            _dummy_solve_result(jax.random.PRNGKey(0)),
            _dummy_solve_result(jax.random.PRNGKey(1)),
            _dummy_solve_result(jax.random.PRNGKey(2)),
            _dummy_solve_result(jax.random.PRNGKey(3)),
        ]
        thicknesses = [1.0, 1.5, 2.0, 2.5]
        expected = scattering.stack_s_matrix(solve_results, thicknesses)
        result = jax.jit(scattering.stack_s_matrix)(solve_results, thicknesses)

        with self.subTest("s11"):
            onp.testing.assert_allclose(result.s11, expected.s11)
        with self.subTest("s12"):
            onp.testing.assert_allclose(result.s12, expected.s12)
        with self.subTest("s21"):
            onp.testing.assert_allclose(result.s21, expected.s21)
        with self.subTest("s22"):
            onp.testing.assert_allclose(result.s22, expected.s22)

    def test_insensitive_to_eigenvalue_ordering(self):
        # Sanity check which validates that the ordering of the eigenvalues and
        # eigenvectors in an internal layer does not impact the final scattering
        # matrix.
        solve_results = _stack_solve_result(jax.random.PRNGKey(0))
        thicknesses = [1.0, 1.5, 2.0, 2.5, 1.0]

        expected = scattering.stack_s_matrix(solve_results, thicknesses)

        q = solve_results[2].eigenvalues
        phi = solve_results[2].eigenvectors
        swapped_q = jnp.concatenate([q[::2], q[1::2]])
        swapped_phi = jnp.concatenate([phi[:, ::2], phi[:, 1::2]], axis=1)

        swapped = dataclasses.replace(
            solve_results[2], eigenvalues=swapped_q, eigenvectors=swapped_phi
        )
        swapped_solve_results = [
            solve_results[0],
            solve_results[1],
            swapped,
            solve_results[3],
            solve_results[4],
        ]
        result = scattering.stack_s_matrix(swapped_solve_results, thicknesses)

        with self.subTest("s11"):
            onp.testing.assert_allclose(result.s11, expected.s11)
        with self.subTest("s12"):
            onp.testing.assert_allclose(result.s12, expected.s12)
        with self.subTest("s21"):
            onp.testing.assert_allclose(result.s21, expected.s21)
        with self.subTest("s22"):
            onp.testing.assert_allclose(result.s22, expected.s22)


class ChangeLayerThicknessTest(unittest.TestCase):
    def test_change_start_layer_thickness(self):
        layer_solve_results = [
            _dummy_solve_result(jax.random.PRNGKey(0)),
            _dummy_solve_result(jax.random.PRNGKey(1)),
            _dummy_solve_result(jax.random.PRNGKey(2)),
            _dummy_solve_result(jax.random.PRNGKey(3)),
        ]
        original = scattering.stack_s_matrix(
            layer_solve_results, layer_thicknesses=[0.5, 0.2, 0.3, 0.2]
        )
        adjusted = scattering.set_start_layer_thickness(original, 0.2)
        expected = scattering.stack_s_matrix(
            layer_solve_results, layer_thicknesses=[0.2, 0.2, 0.3, 0.2]
        )

        onp.testing.assert_allclose(adjusted.s11, expected.s11)
        onp.testing.assert_allclose(adjusted.s21, expected.s21)
        onp.testing.assert_allclose(adjusted.s22, expected.s22)
        onp.testing.assert_allclose(adjusted.s12, expected.s12)
        onp.testing.assert_allclose(
            adjusted.start_layer_thickness, expected.start_layer_thickness
        )
        onp.testing.assert_allclose(
            adjusted.end_layer_thickness, expected.end_layer_thickness
        )

    def test_change_end_layer_thickness(self):
        layer_solve_results = [
            _dummy_solve_result(jax.random.PRNGKey(0)),
            _dummy_solve_result(jax.random.PRNGKey(1)),
            _dummy_solve_result(jax.random.PRNGKey(2)),
            _dummy_solve_result(jax.random.PRNGKey(3)),
        ]
        original = scattering.stack_s_matrix(
            layer_solve_results, layer_thicknesses=[0.2, 0.2, 0.3, 0.5]
        )
        adjusted = scattering.set_end_layer_thickness(original, 0.25)
        expected = scattering.stack_s_matrix(
            layer_solve_results, layer_thicknesses=[0.2, 0.2, 0.3, 0.25]
        )

        onp.testing.assert_allclose(adjusted.s11, expected.s11)
        onp.testing.assert_allclose(adjusted.s21, expected.s21)
        onp.testing.assert_allclose(adjusted.s22, expected.s22)
        onp.testing.assert_allclose(adjusted.s12, expected.s12)
        onp.testing.assert_allclose(
            adjusted.start_layer_thickness, expected.start_layer_thickness
        )
        onp.testing.assert_allclose(
            adjusted.end_layer_thickness, expected.end_layer_thickness
        )



"""An example which computes wavelength-dependent reflection of metallic pillars.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

from typing import Tuple

import jax.numpy as jnp

from fmmax import basis, fields, fmm, layer, scattering, utils

SMatricesInterior = Tuple[
    Tuple[scattering.ScatteringMatrix, scattering.ScatteringMatrix], ...
]

WAVELENGTH_NM: jnp.ndarray = jnp.arange(440, 660, 10)


# Electronvolt-nanometer; can be divided by the wavelength to obtain photon energy.
EV_NM: float = 1239.84198  # hbar * c * 2 * pi / (1 nm * wavelength_nm) in electronvolts


def _to_ev(wavelength_nm: jnp.ndarray) -> jnp.ndarray:
    return EV_NM / wavelength_nm


# Rakic Laurence-Drude model for the permittivity of Ag.
# Rakić et al. 1998, https://doi.org/10.1364/AO.37.005271

F0: float = 0.845
OMEGA_p: float = 9.010  # eV
GAMMA_0: float = 0.048  # eV

F1: float = 0.065
GAMMA_1: float = 3.886  # eV
OMEGA_1: float = 0.816  # eV

F2: float = 0.124
GAMMA_2: float = 0.452  # eV
OMEGA_2: float = 4.481  # eV

F3: float = 0.011
GAMMA_3: float = 0.065  # eV
OMEGA_3: float = 8.185  # eV

F4: float = 0.840
GAMMA_4: float = 0.916  # eV
OMEGA_4: float = 9.083  # eV

F5: float = 5.646
GAMMA_5: float = 2.419  # eV
OMEGA_5: float = 20.29  # eV


def permittivity_ag_drude(wavelength_nm: jnp.ndarray) -> jnp.ndarray:
    """Computes the permittivity of silver for the specified wavelengths."""
    omega = _to_ev(wavelength_nm)
    permittivity = (
        (1 - F0 * OMEGA_p**2 / (omega * (omega + 1j * GAMMA_0)))
        + F1 * OMEGA_p**2 / ((OMEGA_1**2 - omega**2) - 1j * omega * GAMMA_1)
        + F2 * OMEGA_p**2 / ((OMEGA_2**2 - omega**2) - 1j * omega * GAMMA_2)
        + F3 * OMEGA_p**2 / ((OMEGA_3**2 - omega**2) - 1j * omega * GAMMA_3)
        + F4 * OMEGA_p**2 / ((OMEGA_4**2 - omega**2) - 1j * omega * GAMMA_4)
        + F5 * OMEGA_p**2 / ((OMEGA_5**2 - omega**2) - 1j * omega * GAMMA_5)
    )
    return jnp.asarray(permittivity)


def pillar_density(
    pitch_nm: float,
    pillar_diameter_nm: float,
    resolution_nm: float,
) -> jnp.ndarray:
    """Returns the density for a pillar array."""
    x_nm, y_nm = jnp.meshgrid(
        jnp.arange(-pitch_nm / 2, pitch_nm / 2, resolution_nm),
        jnp.arange(-pitch_nm / 2, pitch_nm / 2, resolution_nm),
        indexing="ij",
    )
    distance_nm = jnp.sqrt(x_nm**2 + y_nm**2)
    density = (distance_nm <= pillar_diameter_nm / 2).astype(float)
    return density


def simulate_pillars(
    permittivity_ambient: complex = 1.0 + 0.0j,
    permittivity_planarization: complex = 2.25 + 0.0j,
    wavelength_nm: jnp.ndarray = WAVELENGTH_NM,
    polar_angle: float = 0.0,
    pitch_nm: float = 180.0,
    pillar_diameter_nm: float = 60.0,
    ambient_thickness_nm: float = 1000.0,
    planarization_thickness_nm: float = 20.0,
    pillar_thickness_nm: float = 80.0,
    substrate_thickness_nm: float = 100.0,
    resolution_nm: float = 1.0,
    approximate_num_terms: int = 200,
    truncation: basis.Truncation = basis.Truncation.CIRCULAR,
    formulation: fmm.Formulation = fmm.Formulation.FFT,
) -> Tuple[
    int,
    jnp.ndarray,
    jnp.ndarray,
    Tuple[
        Tuple[layer.LayerSolveResult, ...],
        Tuple[jnp.ndarray, ...],
        SMatricesInterior,
    ],
]:
    """Computes the reflection from an array of circular silver pillars.

    Args:
        permittivity_ambient: The permittivity of the ambient.
        permittivity_planarization: The permittivity of media encapsulating pillars.
        wavelength_nm: The excitation wavelengths, in nanometers.
        polar_angle: The polar angle of the incident plane wave.
        pitch_nm: The pillar pitch, in nanometers.
        pillar_diameter_nm: The diameter of the pillars, in nanometers.
        ambient_thickness_nm: The thickness of the ambient.
        planarization_thickness_nm: The thickness of the planarization layer above
            the pillars.
        pillar_thickness_nm: The thickness of the pillar layer.
        substrate_thickness_nm: The thickness of the substrate.
        resolution_nm: The rasterization resolution for patterned layers.
        approximate_num_terms: The approximate number of terms used in the plane
            wave expansion of the fields.
        truncation: Determines the truncation of the expansion.
        formulation: Specifies the formulation to be used.

    Returns:
        The number of terms in the expansion, the te- and tm-polarized reflection
        coefficients, and intermediate results from the simulation which can be used
        to compute fields. These intermediate results are the layer solve results, the
        layer thicknesses, and the interior scattering matrices.
    """
    density = pillar_density(pitch_nm, pillar_diameter_nm, resolution_nm)
    permittivity_ag = permittivity_ag_drude(wavelength_nm)[:, jnp.newaxis, jnp.newaxis]
    permittivity_pillars = utils.interpolate_permittivity(
        permittivity_solid=permittivity_ag,
        permittivity_void=jnp.asarray(permittivity_planarization),
        density=density[jnp.newaxis, :, :],
    )

    permittivities = [
        jnp.asarray(permittivity_ambient)[jnp.newaxis, jnp.newaxis, jnp.newaxis],
        jnp.asarray(permittivity_planarization)[jnp.newaxis, jnp.newaxis, jnp.newaxis],
        permittivity_pillars,
        permittivity_ag,
    ]
    thicknesses = [
        jnp.asarray(ambient_thickness_nm),
        jnp.asarray(planarization_thickness_nm),
        jnp.asarray(pillar_thickness_nm),
        jnp.asarray(substrate_thickness_nm),
    ]

    in_plane_wavevector = basis.plane_wave_in_plane_wavevector(
        wavelength=wavelength_nm,
        polar_angle=jnp.asarray(polar_angle),
        azimuthal_angle=jnp.zeros(()),
        permittivity=jnp.asarray(permittivity_ambient),
    )
    primitive_lattice_vectors = basis.LatticeVectors(
        u=pitch_nm * basis.X, v=pitch_nm * basis.Y
    )
    expansion = basis.generate_expansion(
        primitive_lattice_vectors=primitive_lattice_vectors,
        approximate_num_terms=approximate_num_terms,
        truncation=truncation,
    )
    layer_solve_results = [
        layer.eigensolve_isotropic_media(
            wavelength=jnp.asarray(wavelength_nm),
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            permittivity=p,
            expansion=expansion,
            formulation=formulation,
        )
        for p in permittivities
    ]

    s_matrices_interior = scattering.stack_s_matrices_interior(
        layer_solve_results, thicknesses
    )

    # The full stack scattering matrix is just the one "below" the final layer.
    s_matrix = s_matrices_interior[-1][0]

    # The reflection coefficient for normally-incident TE-polarized plane
    # waves is the first element of the s21 block of the scattering matrix.
    r_te = s_matrix.s21[..., 0, 0]
    r_tm = s_matrix.s21[..., expansion.num_terms, expansion.num_terms]
    return (
        expansion.num_terms,
        r_te,
        r_tm,
        (tuple(layer_solve_results), tuple(thicknesses), s_matrices_interior),
    )


def compute_fields(
    layer_solve_results: Tuple[layer.LayerSolveResult, ...],
    thicknesses: Tuple[jnp.ndarray, ...],
    s_matrices_interior: SMatricesInterior,
    resolution_nm: float,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Computes the electric and magnetic fields in the simulation volume.

    Args:
        layer_solve_results: The results of the eigensolve for each layer.
        thicknesses: The thickness of each layer.
        s_matrices_interior: The interior scattering matrices.
        resolution_nm: The resolution for the grid on which fields are computed.

    Returns:
        The electric and magnetic fields, each with shape `(3, num_wavelengths, nx,
        ny, nz)`. The leading axis is for the field direction, i.e. x, y, and z.
    """

    # Excite with a TE_polarized plane wave.
    num_terms = layer_solve_results[0].expansion.num_terms
    forward_amplitude_0_start = jnp.zeros((1, 2 * num_terms, 1), dtype=complex)
    forward_amplitude_0_start = forward_amplitude_0_start.at[:, 0, 0].set(1)
    backward_amplitude_N_end = jnp.zeros_like(forward_amplitude_0_start)

    # Compute the wave amplitudes within each layer.
    amplitudes_interior = fields.stack_amplitudes_interior(
        s_matrices_interior, forward_amplitude_0_start, backward_amplitude_N_end
    )

    efields, hfields, (x, y, z) = fields.stack_fields_3d_auto_grid(
        amplitudes_interior=amplitudes_interior,
        layer_solve_results=layer_solve_results,
        layer_thicknesses=thicknesses,
        grid_spacing=resolution_nm,
        num_unit_cells=(1, 1),
    )
    return efields, hfields


if __name__ == "__main__":
    print(simulate_pillars())



"""Tests for `fmmax.farfield`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import functools
import unittest

import jax
import jax.numpy as jnp
import numpy as onp
import parameterized

from fmmax import basis, farfield, fields, fmm, layer, scattering, sources, utils


class FarfieldProfileTest(unittest.TestCase):
    def test_dipole_farfield_matches_analytical_calculation(self):
        # Calculate the farfield for a dipole in vacuum, and compare to an analytical result.
        wavelength = jnp.asarray(0.63)
        primitive_lattice_vectors = basis.LatticeVectors(basis.X, basis.Y)
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=primitive_lattice_vectors,
            approximate_num_terms=50,
            truncation=basis.Truncation.CIRCULAR,
        )
        in_plane_wavevector = basis.brillouin_zone_in_plane_wavevector(
            (7, 7), primitive_lattice_vectors
        )

        dipole = sources.dirac_delta_source(
            location=jnp.asarray([[0.5, 0.5]]),
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            expansion=expansion,
        )
        zeros = jnp.zeros_like(dipole)
        jx = jnp.concatenate([dipole, zeros, zeros], axis=-1)
        jy = jnp.concatenate([zeros, dipole, zeros], axis=-1)
        jz = jnp.concatenate([zeros, zeros, dipole], axis=-1)

        layer_solve_result = layer.eigensolve_isotropic_media(
            permittivity=jnp.asarray([[1.0]]),
            wavelength=jnp.asarray(wavelength),
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            expansion=expansion,
            formulation=fmm.Formulation.FFT,
        )
        s_matrix_before_source = scattering.stack_s_matrix([layer_solve_result], [1.0])
        s_matrix_after_source = s_matrix_before_source
        bwd_amplitude_ambient_end, *_ = sources.amplitudes_for_source(
            jx=jx,
            jy=jy,
            jz=jz,
            s_matrix_before_source=s_matrix_before_source,
            s_matrix_after_source=s_matrix_after_source,
        )
        forward_flux, backward_flux = fields.directional_poynting_flux(
            forward_amplitude=jnp.zeros_like(bwd_amplitude_ambient_end),
            backward_amplitude=bwd_amplitude_ambient_end,
            layer_solve_result=layer_solve_result,
        )
        (
            polar_angle,
            azimuthal_angle,
            solid_angle,
            farfield_flux,
        ) = farfield.farfield_profile(
            flux=-backward_flux,
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            expansion=expansion,
            brillouin_grid_axes=(0, 1),
        )
        # Sum the farfield contributions from the two polarizations, and normalize.
        farfield_flux = jnp.sum(farfield_flux, axis=-2)
        mask = ~jnp.isnan(farfield_flux)
        farfield_flux /= jnp.amax(farfield_flux[mask])

        # Compute an analytical farfield, which has amplitude `sin(theta)`, when
        # theta is the angle with respect to the polarization axis.
        x = jnp.sin(polar_angle) * jnp.cos(azimuthal_angle)
        y = jnp.sin(polar_angle) * jnp.sin(azimuthal_angle)
        z_flux = jnp.sin(polar_angle) ** 2
        x_flux = jnp.sin(jnp.arccos(x)) ** 2
        y_flux = jnp.sin(jnp.arccos(y)) ** 2
        analytical_farfield_flux = jnp.stack([x_flux, y_flux, z_flux], axis=-1)

        onp.testing.assert_allclose(
            farfield_flux[mask],
            analytical_farfield_flux[mask],
            rtol=1e-5,
        )

    @parameterized.parameterized.expand(
        (
            [(1, 1), (0, 1), jnp.pi / 2],
            [(4, 5), (0, 1), jnp.pi / 2],
            [(4, 5), (-4, -3), jnp.pi / 2],
            [(2, 4, 5), (-4, -3), jnp.pi / 2],
            [(4, 5, 2), (-5, -4), jnp.pi / 2],
            [(1, 4, 5, 2, 1), (-6, -5), jnp.pi / 2],
            [(4, 5), (0, 1), 0.8 * jnp.pi / 2],  # Integrals over fraction of farfield.
            [(4, 5), (0, 1), 0.5 * jnp.pi / 2],
            [(4, 5), (0, 1), 0.2 * jnp.pi / 2],
        )
    )
    def test_farfield_integral_in_different_domains(
        self, batch_shape, brillouin_grid_axes, polar_angle_cutoff
    ):
        # Checks that the farfield can be computed for a variety of batch
        # shapes and brillouin zone axes. Also checks that the integral of
        # flux in k space and in the angular domain agree.
        absolute_axes = utils.absolute_axes(brillouin_grid_axes, len(batch_shape) + 2)

        # Compute shapes of dummy variables.
        wavelength_shape = tuple(
            [1 if i in absolute_axes else d for i, d in enumerate(batch_shape)]
        )
        bz_grid_shape = tuple(
            [d for i, d in enumerate(batch_shape) if i in absolute_axes]
        )

        # Generate dummy variable shapes.
        wavelength = 1 + jnp.arange(jnp.prod(jnp.asarray(wavelength_shape)))
        wavelength = wavelength.reshape(wavelength_shape)
        primitive_lattice_vectors = basis.LatticeVectors(basis.X, basis.Y)
        in_plane_wavevector = basis.brillouin_zone_in_plane_wavevector(
            brillouin_grid_shape=bz_grid_shape,
            primitive_lattice_vectors=primitive_lattice_vectors,
        )
        in_plane_wavevector = jnp.expand_dims(
            in_plane_wavevector,
            [i for i in range(len(batch_shape)) if i not in absolute_axes],
        )

        expansion = basis.generate_expansion(
            primitive_lattice_vectors=primitive_lattice_vectors,
            approximate_num_terms=100,
            truncation=basis.Truncation.CIRCULAR,
        )
        transverse_wavevectors = basis.transverse_wavevectors(
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            expansion=expansion,
        )

        kx = transverse_wavevectors[..., 0]
        ky = transverse_wavevectors[..., 1]
        kt = jnp.sqrt(kx**2 + ky**2)
        kz = jnp.sqrt((2 * jnp.pi / wavelength[..., jnp.newaxis]) ** 2 - kt**2)
        kt_cutoff = jnp.sin(polar_angle_cutoff) * 2 * jnp.pi / wavelength

        # Construct a dummy flux which is 1 for all orders that have nonzero kz,
        # i.e. those which actually carry power into the farfield.
        flux = jnp.concatenate([jnp.where(jnp.isnan(kz), 0, 1)] * 2, axis=-1)
        flux = flux[..., jnp.newaxis]

        # Compute the flux within the specified angular cone.
        k_space_mask = kt < kt_cutoff[..., jnp.newaxis]
        k_space_mask = jnp.concatenate([k_space_mask, k_space_mask], axis=-1)
        k_space_mask = k_space_mask[..., jnp.newaxis]

        # Integrate the flux over k-space. The differential area element in
        # k-space is independent of k.
        integrated_flux_k_space = jnp.sum(
            jnp.where(k_space_mask, flux, 0), axis=brillouin_grid_axes + (-2,)
        )

        # Compute the farfield angles, solid angle, and associated flux.
        (
            polar_angle,
            azimuthal_angle,
            solid_angle,
            farfield_flux,
        ) = farfield.farfield_profile(
            flux=flux,
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            expansion=expansion,
            brillouin_grid_axes=brillouin_grid_axes,
        )
        farfield_flux = jnp.where(jnp.isnan(farfield_flux), 0, farfield_flux)
        solid_angle = solid_angle[..., jnp.newaxis, jnp.newaxis]

        angle_mask = polar_angle < polar_angle_cutoff
        angle_mask = jnp.where(jnp.isnan(angle_mask), False, angle_mask)
        angle_mask = angle_mask[..., jnp.newaxis, jnp.newaxis]

        # Integrate the flux in the angular domain. The differential element
        # is `solid_angle`, i.e. the area of the cell associated with each
        # element in `farfield_flux`.
        integrated_flux_spherical_coords = jnp.sum(
            jnp.where(angle_mask, farfield_flux * solid_angle, 0), axis=(-4, -3, -2)
        )
        onp.testing.assert_allclose(
            integrated_flux_spherical_coords, integrated_flux_k_space
        )


class IntegratedFluxTest(unittest.TestCase):
    def test_resize(self):
        # Directly tests resizing, as this can fail if some GPU libraries are missing.
        arr = jnp.ones((10, 10))
        upsampled = jax.image.resize(arr, shape=(30, 30), method="linear")
        self.assertSequenceEqual(upsampled.shape, (30, 30))

    @parameterized.parameterized.expand(
        (
            [(4, 4), (0, 1), 1, 1],
            [(2, 4, 4), (1, 2), 1, 1],
            [(4, 4, 2), (0, 1), 1, 1],
            [(4, 4, 2), (0, 1), 5, 1],
            [(4, 2, 4), (0, 2), 5, 1],
            [(4, 4, 2), (0, 1), 1, 2],
        )
    )
    def test_integrated_flux_matches_direct_calculation(
        self, batch_shape, brillouin_grid_axes, num_sources, upsample_factor
    ):
        # Checks that the calculation of integrated flux "directly" matches that
        # when calcuated by first computing weights, and then taking the inner
        # product.

        # Dummy values to test the farfield calculation.
        primitive_lattice_vectors = basis.LatticeVectors(basis.X * 3, basis.Y)
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=primitive_lattice_vectors,
            approximate_num_terms=100,
            truncation=basis.Truncation.CIRCULAR,
        )
        flux = jax.random.uniform(
            jax.random.PRNGKey(0),
            batch_shape + (2 * expansion.num_terms, num_sources),
        )
        wavelength = jax.random.uniform(
            jax.random.PRNGKey(1),
            [(1 if i in brillouin_grid_axes else d) for i, d in enumerate(batch_shape)],
        )
        in_plane_wavevector = basis.brillouin_zone_in_plane_wavevector(
            brillouin_grid_shape=tuple([batch_shape[i] for i in brillouin_grid_axes]),
            primitive_lattice_vectors=primitive_lattice_vectors,
        )
        in_plane_wavevector = jnp.expand_dims(
            in_plane_wavevector,
            axis=[i for i in range(len(batch_shape)) if i not in brillouin_grid_axes],
        )

        def angle_bounds_fn(polar_angle, azimuthal_angle):
            del azimuthal_angle
            return jnp.full(polar_angle.shape, True)

        # Calculate the integrated flux by the default method, which first
        # computes weights and then does the integration by taking the weighted
        # sum of flux.
        integrated_flux = farfield.integrated_flux(
            flux=flux,
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            expansion=expansion,
            brillouin_grid_axes=brillouin_grid_axes,
            angle_bounds_fn=angle_bounds_fn,
            upsample_factor=upsample_factor,
        )

        # Compute the integrated flux directly.
        integrated_flux_direct = farfield._integrated_flux_upsampled(
            flux=flux,
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            expansion=expansion,
            brillouin_grid_axes=brillouin_grid_axes,
            angle_bounds_fn=angle_bounds_fn,
            upsample_factor=upsample_factor,
        )
        onp.testing.assert_allclose(integrated_flux, integrated_flux_direct, rtol=1e-6)

    def test_upsample_scale(self):
        # Check that scaling of the integrated power when using upsampling is correct.
        primitive_lattice_vectors = basis.LatticeVectors(basis.X * 3, basis.Y)
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=primitive_lattice_vectors,
            approximate_num_terms=100,
            truncation=basis.Truncation.CIRCULAR,
        )
        flux = jax.random.uniform(
            jax.random.PRNGKey(0),
            (4, 4, 2 * expansion.num_terms, 1),
        )
        wavelength = jnp.asarray(0.63)
        in_plane_wavevector = basis.brillouin_zone_in_plane_wavevector(
            brillouin_grid_shape=(4, 4),
            primitive_lattice_vectors=primitive_lattice_vectors,
        )

        def angle_bounds_fn(polar_angle, azimuthal_angle):
            del azimuthal_angle
            return jnp.full(polar_angle.shape, True)

        integrated_flux_fn = functools.partial(
            farfield.integrated_flux,
            flux=flux,
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            expansion=expansion,
            brillouin_grid_axes=(0, 1),
            angle_bounds_fn=angle_bounds_fn,
        )
        integrated_flux_no_upsample = integrated_flux_fn(upsample_factor=1)
        integrated_flux_upsample = integrated_flux_fn(upsample_factor=10)
        onp.testing.assert_allclose(
            integrated_flux_no_upsample, integrated_flux_upsample
        )

    def test_all_angles_matches_total_flux(self):
        wavelength = jnp.asarray(0.63)
        primitive_lattice_vectors = basis.LatticeVectors(basis.X, basis.Y)
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=primitive_lattice_vectors,
            approximate_num_terms=50,
            truncation=basis.Truncation.CIRCULAR,
        )
        in_plane_wavevector = basis.brillouin_zone_in_plane_wavevector(
            (7, 7), primitive_lattice_vectors
        )

        dipole = sources.dirac_delta_source(
            location=jnp.asarray([[0.5, 0.5]]),
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            expansion=expansion,
        )
        zeros = jnp.zeros_like(dipole)
        jx = jnp.concatenate([dipole, zeros, zeros], axis=-1)
        jy = jnp.concatenate([zeros, dipole, zeros], axis=-1)
        jz = jnp.concatenate([zeros, zeros, dipole], axis=-1)

        layer_solve_result = layer.eigensolve_isotropic_media(
            permittivity=jnp.asarray([[1.0]]),
            wavelength=jnp.asarray(wavelength),
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            expansion=expansion,
            formulation=fmm.Formulation.FFT,
        )
        s_matrix_before_source = scattering.stack_s_matrix([layer_solve_result], [1.0])
        s_matrix_after_source = s_matrix_before_source
        (
            bwd_amplitude_ambient_end,
            fwd_amplitude_before_start,
            bwd_amplitude_before_end,
            fwd_amplitude_after_start,
            bwd_amplitude_after_end,
            fwd_amplitude_substrate_start,
        ) = sources.amplitudes_for_source(
            jx=jx,
            jy=jy,
            jz=jz,
            s_matrix_before_source=s_matrix_before_source,
            s_matrix_after_source=s_matrix_after_source,
        )
        forward_flux, backward_flux = fields.directional_poynting_flux(
            forward_amplitude=jnp.zeros_like(bwd_amplitude_ambient_end),
            backward_amplitude=bwd_amplitude_ambient_end,
            layer_solve_result=layer_solve_result,
        )
        (
            polar_angle,
            azimuthal_angle,
            solid_angle,
            farfield_flux,
        ) = farfield.farfield_profile(
            flux=-backward_flux,
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            expansion=expansion,
            brillouin_grid_axes=(0, 1),
        )

        integrated_flux = farfield.integrated_flux(
            flux=-backward_flux,
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            expansion=expansion,
            brillouin_grid_axes=(0, 1),
            angle_bounds_fn=lambda polar_angle, _: jnp.full(polar_angle.shape, True),
            upsample_factor=10,
        )
        onp.testing.assert_allclose(
            jnp.sum(-backward_flux),
            jnp.sum(integrated_flux),
            rtol=1e-6,
        )


class UnflattenTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        (
            ([[1, 2, 3], [4, 5, 6]],),
            ([[[1, 2, 3], [4, 5, 6]]],),
            ([[[1, 2, 3], [4, 5, 6]], [[7, 8, 9], [10, 11, 12]]],),
            ([[[[1, 2, 3], [4, 5, 6]]], [[[7, 8, 9], [10, 11, 12]]]],),
        )
    )
    def test_unflatten_matches_expected(self, bz_data):
        bz_data = jnp.asarray(bz_data)

        expansion = basis.generate_expansion(
            primitive_lattice_vectors=basis.LatticeVectors(basis.X * 3, basis.Y),
            approximate_num_terms=100,
            truncation=basis.Truncation.PARALLELOGRAMIC,
        )

        # Create a dummy array to be unstacked, where all the Fourier orders for a
        # given Brillouin zone point just take the value in `bz_data`.
        data = jnp.broadcast_to(
            bz_data[..., jnp.newaxis], bz_data.shape + (expansion.num_terms,)
        )
        unstacked = farfield.unflatten(data, expansion)

        i_min = onp.amin(expansion.basis_coefficients[:, 0])
        i_max = onp.amax(expansion.basis_coefficients[:, 0])
        j_min = onp.amin(expansion.basis_coefficients[:, 1])
        j_max = onp.amax(expansion.basis_coefficients[:, 1])

        batch_shape = bz_data.shape[:-2]
        expected_shape = batch_shape + (
            bz_data.shape[-2] * (i_max - i_min + 1),
            bz_data.shape[-1] * (j_max - j_min + 1),
        )
        self.assertSequenceEqual(unstacked.shape, expected_shape)

        # The unflattening of dummy data is just the tiling of the original `bz_data`.
        expected = onp.tile(
            bz_data,
            (1,) * (bz_data.ndim - 2) + (i_max - i_min + 1, j_max - j_min + 1),
        )
        onp.testing.assert_array_equal(expected, unstacked)

    @parameterized.parameterized.expand(
        (
            [(4, 5), (0, 1)],
            [(2, 4, 5), (1, 2)],
            [(1, 3, 2, 4), (1, 2)],
            [(1, 3, 2, 4), (-5, -4)],
        )
    )
    def test_unflatten_flux_matches_expected(self, batch_shape, bz_axes):
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=basis.LatticeVectors(basis.X, basis.Y * 3),
            approximate_num_terms=100,
            truncation=basis.Truncation.PARALLELOGRAMIC,
        )

        i_min = onp.amin(expansion.basis_coefficients[:, 0])
        i_max = onp.amax(expansion.basis_coefficients[:, 0])
        j_min = onp.amin(expansion.basis_coefficients[:, 1])
        j_max = onp.amax(expansion.basis_coefficients[:, 1])
        nkx = i_max - i_min + 1
        nky = j_max - j_min + 1

        # Create a dummy flux array with batch dimensions and four sources.
        # This array is missing the Fourier order axis; we will broadcast and
        # reshape so that all Fourier orders for a given batch element and
        # polarization are equal.
        num_sources = 4
        batch_flux = onp.arange(onp.prod(batch_shape) * 2 * num_sources)
        batch_flux = batch_flux.reshape(batch_shape + (2, num_sources))

        # Broadcast and reshape to have the proper shape of a flux array.
        flux = onp.broadcast_to(
            batch_flux[..., jnp.newaxis, :],
            batch_shape + (2, expansion.num_terms, num_sources),
        )
        flux = flux.reshape(batch_shape + (2 * expansion.num_terms, num_sources))

        unstacked = farfield.unflatten_flux(flux, expansion, bz_axes)

        # Compute expected shape of the unflattend flux array.
        bz_axes = tuple([b % flux.ndim for b in bz_axes])
        unstacked_batch_shape = tuple(
            [d for i, d in enumerate(batch_shape) if i not in bz_axes]
        )
        expected_shape = unstacked_batch_shape + (
            nkx * batch_shape[bz_axes[0]],
            nky * batch_shape[bz_axes[1]],
            2,
            num_sources,
        )
        self.assertSequenceEqual(unstacked.shape, expected_shape)

        # Compute the expected unflattened flux array.
        transposed_axes = (
            tuple([i for i in range(batch_flux.ndim - 2) if i not in bz_axes])
            + bz_axes
            + (flux.ndim - 2, flux.ndim - 1)
        )
        transposed_batch_flux = onp.transpose(batch_flux, axes=transposed_axes)
        expected = onp.tile(
            transposed_batch_flux, (1,) * len(unstacked_batch_shape) + (nkx, nky, 1, 1)
        )
        onp.testing.assert_array_equal(expected, unstacked)

    @parameterized.parameterized.expand(
        (
            [(4, 5), (0, 1)],
            [(2, 4, 5), (1, 2)],
            [(1, 3, 2, 4), (1, 2)],
            [(1, 3, 2, 4), (-5, -4)],
        )
    )
    def test_unflatten_wavevector_matches_expected(self, batch_shape, bz_axes):
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=basis.LatticeVectors(basis.X, basis.Y * 3),
            approximate_num_terms=100,
            truncation=basis.Truncation.PARALLELOGRAMIC,
        )

        i_min = onp.amin(expansion.basis_coefficients[:, 0])
        i_max = onp.amax(expansion.basis_coefficients[:, 0])
        j_min = onp.amin(expansion.basis_coefficients[:, 1])
        j_max = onp.amax(expansion.basis_coefficients[:, 1])
        nkx = i_max - i_min + 1
        nky = j_max - j_min + 1

        # Create a dummy transverse wavevector array with batch dimensions.
        # This array is missing the Fourier order axis; we will broadcast and
        # reshape so that all Fourier orders for a given batch element are equal.
        batch_wavevectors = onp.arange(onp.prod(batch_shape) * 2)
        batch_wavevectors = batch_wavevectors.reshape(batch_shape + (2,))

        # Broadcast and reshape to have the proper shape of a wavevector array.
        wavevectors = onp.broadcast_to(
            batch_wavevectors[..., jnp.newaxis, :],
            batch_shape + (expansion.num_terms, 2),
        )

        unstacked = farfield.unflatten_transverse_wavevectors(
            wavevectors, expansion, bz_axes
        )

        # Compute expected shape of the unflattend flux array.
        bz_axes = tuple([b % wavevectors.ndim for b in bz_axes])
        unstacked_batch_shape = tuple(
            [d for i, d in enumerate(batch_shape) if i not in bz_axes]
        )
        expected_shape = unstacked_batch_shape + (
            nkx * batch_shape[bz_axes[0]],
            nky * batch_shape[bz_axes[1]],
            2,
        )
        self.assertSequenceEqual(unstacked.shape, expected_shape)

        # Compute the expected unflattened transverse wavevector array.
        transposed_axes = (
            tuple([i for i in range(batch_wavevectors.ndim - 1) if i not in bz_axes])
            + bz_axes
            + (batch_wavevectors.ndim - 1,)
        )
        transposed_flux = onp.transpose(batch_wavevectors, axes=transposed_axes)
        expected = onp.tile(
            transposed_flux, (1,) * len(unstacked_batch_shape) + (nkx, nky, 1)
        )
        onp.testing.assert_array_equal(expected, unstacked)

    @parameterized.parameterized.expand(
        (
            [(1, 1), basis.LatticeVectors(basis.X, basis.Y)],
            [(3, 4), basis.LatticeVectors(basis.X, basis.Y)],
            [(1, 1), basis.LatticeVectors(basis.X, -basis.Y)],
            [(3, 4), basis.LatticeVectors(basis.X, -basis.Y)],
            [(1, 1), basis.LatticeVectors(basis.X, basis.Y * 3)],
            [(3, 4), basis.LatticeVectors(basis.X, basis.Y * 3)],
            [(1, 1), basis.LatticeVectors(basis.X + basis.Y, basis.X - basis.Y)],
            [(3, 4), basis.LatticeVectors(basis.X + basis.Y, basis.X - basis.Y)],
        )
    )
    def test_unstacked_wavevectors_are_sorted(
        self, bz_shape, primitive_lattice_vectors
    ):
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=primitive_lattice_vectors,
            approximate_num_terms=100,
            truncation=basis.Truncation.PARALLELOGRAMIC,
        )
        in_plane_wavevector = basis.brillouin_zone_in_plane_wavevector(
            brillouin_grid_shape=bz_shape,
            primitive_lattice_vectors=primitive_lattice_vectors,
        )
        transverse_wavevectors = basis.transverse_wavevectors(
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            expansion=expansion,
        )
        unflattened_transverse_wavevectors = farfield.unflatten_transverse_wavevectors(
            transverse_wavevectors=transverse_wavevectors,
            expansion=expansion,
            brillouin_grid_axes=(0, 1),
        )
        kx = unflattened_transverse_wavevectors[..., 0]
        ky = unflattened_transverse_wavevectors[..., 1]
        self.assertEqual(kx.ndim, 2)
        dkx = kx[1:, :] - kx[:-1, :]
        dky = ky[:, 1:] - ky[:, :-1]
        # The wavevectors should be monotonically increasing or decreasing,
        # depending upon the choice of basis vectors.
        self.assertTrue(onp.all(dkx > 0) or onp.all(dkx < 0))
        self.assertTrue(onp.all(dky > 0) or onp.all(dky < 0))



"""A photonic crystal with Gaussian beam excitation and internal source.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import functools
from typing import Tuple

import jax.numpy as jnp
import matplotlib.pyplot as plt  # type: ignore[import]
import numpy as onp
from skimage import measure  # type: ignore[import]

from fmmax import basis, beams, fields, fmm, layer, scattering, sources

PERMITTIVITY_AMBIENT: complex = (1.0 + 0.0j) ** 2
PERMITTIVITY_SLAB: complex = (1.5 + 0.0j) ** 2
THICKNESS_AMBIENT: float = 2.0
THICKNESS_SLAB: float = 0.8
PITCH: float = 1.0
DIAMETER: float = 0.7
RESOLUTION: float = 0.01
RESOLUTION_FIELDS: float = 0.01
WAVELENGTH: float = 0.63
APPROXIMATE_NUM_TERMS: int = 50
BRILLOUIN_GRID_SHAPE: Tuple[int, int] = (9, 9)


def simulate_crystal_with_internal_source(
    permittivity_ambient: complex = PERMITTIVITY_AMBIENT,
    permittivity_slab: complex = PERMITTIVITY_SLAB,
    thickness_ambient: float = THICKNESS_AMBIENT,
    thickness_slab: float = THICKNESS_SLAB,
    pitch: float = PITCH,
    diameter: float = DIAMETER,
    resolution: float = RESOLUTION,
    resolution_fields: float = RESOLUTION_FIELDS,
    wavelength: float = WAVELENGTH,
    approximate_num_terms: int = APPROXIMATE_NUM_TERMS,
    brillouin_grid_shape: Tuple[int, int] = BRILLOUIN_GRID_SHAPE,
) -> Tuple[
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],  # (ex, ey, ez)
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],  # (hx, hy, hz)
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],  # (x, y, z)
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],  # (xy, xz, yz) cross sections
]:
    """Simulates a dipole source inside a photonic crystal slab.

    The crystal has a square unit cell with circular holes, having cross section
    and dipole position as illustrated below. The dipole is located the lower-left
    corner of the unit cell centered in the supercell defined by the Brillouin grid
    shape. The dipole is x-oriented and centered vertically within the photonic
    crystal slab.
                     ________________
                    |                |
                    |XX            XX|
                    |XXXX        XXXX|
                    |XXXX        XXXX|
                    |XX            XX|
        x-dipole -> o________________|

    Args:
        permittivity_ambient: Permittivity of the region above and below the slab, and
            of the holes in the slab.
        permittivity_slab: Permittivity of the slab.
        thickness_ambient: Thickness of the ambient layers above and below the slab.
        thickness_slab: Thickness of the photonic crystal slab.
        pitch: The unit cell pitch.
        diameter: The diameter of the holes in the photonic crystal.
        resolution: The size of a pixel in permittivity arrays.
        resolution_fields: The size of a pixel in field arrays.
        wavelength: The wavelength, of the dipole emission.
        approximate_num_terms: The number of terms in the Fourier expansion.
        brillouin_grid_shape: The shape of the grid used for Brillouin zone integration.

    Returns:
        The electric and magnetic fields, and the grid coordinates, `((ex, ey, ez),
        (hx, hy, hz), (x, y, z))`. The fields are returned for an xz slice centered
        on the dipole.
    """
    thickness_ambient_ = jnp.asarray(thickness_ambient)
    thickness_slab_ = jnp.asarray(thickness_slab)
    del thickness_ambient, thickness_slab

    primitive_lattice_vectors = basis.LatticeVectors(
        u=pitch * basis.X, v=pitch * basis.Y
    )
    expansion = basis.generate_expansion(
        primitive_lattice_vectors=primitive_lattice_vectors,
        approximate_num_terms=approximate_num_terms,
        truncation=basis.Truncation.CIRCULAR,
    )

    # Brillouin zone integration creates a batch of in-plane wavevectors which are
    # distributed throughout the first Brillouin zone.
    in_plane_wavevector = basis.brillouin_zone_in_plane_wavevector(
        brillouin_grid_shape, primitive_lattice_vectors
    )
    assert in_plane_wavevector.shape[-1] == 2
    assert in_plane_wavevector.ndim == 3

    eigensolve = functools.partial(
        layer.eigensolve_isotropic_media,
        wavelength=jnp.asarray(wavelength),
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
        formulation=fmm.Formulation.FFT,
    )

    mask = unit_cell_pattern(pitch, diameter, resolution)
    permittivity_crystal = jnp.where(mask, permittivity_slab, permittivity_slab)
    solve_result_crystal = eigensolve(permittivity=permittivity_crystal)
    solve_result_ambient = eigensolve(
        permittivity=jnp.asarray(permittivity_ambient)[jnp.newaxis, jnp.newaxis]
    )

    # First, we model a dipole inside the photonic crystal. For this, we must break
    # the stack into two, and compute scattering matrices for the stacks above and
    # below the plane containing the dipole. Since we want to visualize fields, we
    # also need the interior scattering matrices.
    s_matrices_interior_before_source = scattering.stack_s_matrices_interior(
        layer_solve_results=[solve_result_ambient, solve_result_crystal],
        layer_thicknesses=[thickness_ambient_, thickness_slab_ / 2],
    )
    s_matrices_interior_after_source = scattering.stack_s_matrices_interior(
        layer_solve_results=[solve_result_crystal, solve_result_ambient],
        layer_thicknesses=[thickness_slab_ / 2, thickness_ambient_],
    )
    # Extract the scattering matrices relating fields at the two ends of each substack.
    s_matrix_before_source = s_matrices_interior_before_source[-1][0]
    s_matrix_after_source = s_matrices_interior_after_source[-1][0]

    # Generate the Fourier representation of a point dipole.
    dipole_x = pitch * brillouin_grid_shape[0] // 2
    dipole_y = pitch * brillouin_grid_shape[1] // 2
    dipole = sources.dirac_delta_source(
        location=jnp.asarray([[dipole_x, dipole_y]]),
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
    )
    # Compute backward eigenmode amplitudes at the end of the layer before the
    # source, and the forward amplitudes the start of the layer after the source.
    (
        _,
        _,
        bwd_amplitude_before_end,
        fwd_amplitude_after_start,
        _,
        _,
    ) = sources.amplitudes_for_source(
        jx=dipole,
        jy=jnp.zeros_like(dipole),
        jz=jnp.zeros_like(dipole),
        s_matrix_before_source=s_matrix_before_source,
        s_matrix_after_source=s_matrix_after_source,
    )

    # Compute the fields inside the structure.
    amplitudes_interior = fields.stack_amplitudes_interior_with_source(
        s_matrices_interior_before_source=s_matrices_interior_before_source,
        s_matrices_interior_after_source=s_matrices_interior_after_source,
        backward_amplitude_before_end=bwd_amplitude_before_end,
        forward_amplitude_after_start=fwd_amplitude_after_start,
    )
    # Coordinates where fields are to be evaluated.
    x = jnp.arange(0, pitch * brillouin_grid_shape[0], resolution_fields)
    y = jnp.ones_like(x) * pitch * brillouin_grid_shape[1] // 2
    (ex, ey, ez), (hx, hy, hz), (x, y, z) = fields.stack_fields_3d_on_coordinates(
        amplitudes_interior=amplitudes_interior,
        layer_solve_results=[
            solve_result_ambient,
            solve_result_crystal,
            solve_result_crystal,
            solve_result_ambient,
        ],
        layer_thicknesses=[
            thickness_ambient_,
            thickness_slab_ / 2,
            thickness_slab_ / 2,
            thickness_ambient_,
        ],
        layer_znum=[
            int(thickness_ambient_ / resolution_fields),
            int(thickness_slab_ / resolution_fields / 2),
            int(thickness_slab_ / resolution_fields / 2),
            int(thickness_ambient_ / resolution_fields),
        ],
        x=x,
        y=y,
    )

    # Perform the Brillouin zone integration by averaging over the Brillouin zone
    # grid batch axes.
    ex, ey, ez, hx, hy, hz = [
        jnp.mean(field, axis=(0, 1)) for field in (ex, ey, ez, hx, hy, hz)
    ]

    # Compute some cross sections for visualizing the structure.
    section_xy, section_xz, section_yz = crystal_cross_sections(
        thickness_ambient=float(thickness_ambient_),
        thickness_slab=float(thickness_slab_),
        pitch=pitch,
        diameter=diameter,
        resolution=resolution,
        num_unit_cells=brillouin_grid_shape,
    )
    return (ex, ey, ez), (hx, hy, hz), (x, y, z), (section_xy, section_xz, section_yz)


def simulate_crystal_with_gaussian_beam(
    polar_angle: float = 0.15 * jnp.pi,
    azimuthal_angle: float = 0.0,
    polarization_angle: float = 0.0,
    beam_waist: float = 1.0,
    beam_focus_offset: float = 0.0,
    permittivity_ambient: complex = PERMITTIVITY_AMBIENT,
    permittivity_slab: complex = PERMITTIVITY_SLAB,
    thickness_ambient: float = THICKNESS_AMBIENT,
    thickness_slab: float = THICKNESS_SLAB,
    pitch: float = PITCH,
    diameter: float = DIAMETER,
    resolution: float = RESOLUTION,
    resolution_fields: float = RESOLUTION_FIELDS,
    wavelength: float = WAVELENGTH,
    approximate_num_terms: int = APPROXIMATE_NUM_TERMS,
    brillouin_grid_shape: Tuple[int, int] = BRILLOUIN_GRID_SHAPE,
) -> Tuple[
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],  # (ex, ey, ez)
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],  # (hx, hy, hz)
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],  # (x, y, z)
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],  # (xy, xz, yz) cross sections
]:
    """Simulates a Gaussian beam incident on photonic crystal slab.

    The crystal has a square unit cell with circular holes as illustrated below.
                     ________________
                    |                |
                    |XX            XX|
                    |XXXX        XXXX|
                    |XXXX        XXXX|
                    |XX            XX|
                    |________________|

    Args:
        polar_angle: The polar angle of the incident beam.
        azimuthal_angle: The azimuthal angle of the incident beam.
        polarization_angle: The angle giving the polarization rotation about the
            propagation axis.
        beam_waist: The Gaussian beam waist.
        beam_focus_offset: The offset of the Gaussian beam focus from the top of the
            photonic crystal slab.
        permittivity_ambient: Permittivity of the region above and below the slab, and
            of the holes in the slab.
        permittivity_slab: Permittivity of the slab.
        thickness_ambient: Thickness of the ambient layers above and below the slab.
        thickness_slab: Thickness of the photonic crystal slab.
        pitch: The unit cell pitch.
        diameter: The diameter of the holes in the photonic crystal.
        resolution: The size of a pixel in permittivity arrays.
        resolution_fields: The size of a pixel in field arrays.
        wavelength: The wavelength, of the dipole emission.
        approximate_num_terms: The number of terms in the Fourier expansion.
        brillouin_grid_shape: The shape of the grid used for Brillouin zone integration.

    Returns:
        The electric and magnetic fields, and the grid coordinates, `((ex, ey, ez),
        (hx, hy, hz), (x, y, z))`. The fields are returned for an xz slice centered
        on the incident beam.
    """
    thickness_ambient_ = jnp.asarray(thickness_ambient)
    thickness_slab_ = jnp.asarray(thickness_slab)
    del thickness_ambient, thickness_slab

    primitive_lattice_vectors = basis.LatticeVectors(
        u=pitch * basis.X, v=pitch * basis.Y
    )
    expansion = basis.generate_expansion(
        primitive_lattice_vectors=primitive_lattice_vectors,
        approximate_num_terms=approximate_num_terms,
        truncation=basis.Truncation.CIRCULAR,
    )

    # Brillouin zone integration creates a batch of in-plane wavevectors which are
    # distributed throughout the first Brillouin zone. We shift the expansion so
    # that it is centered on the direction of the incident beam.
    in_plane_wavevector = basis.brillouin_zone_in_plane_wavevector(
        brillouin_grid_shape, primitive_lattice_vectors
    )
    in_plane_wavevector += basis.plane_wave_in_plane_wavevector(
        wavelength=jnp.asarray(wavelength),
        polar_angle=jnp.asarray(polar_angle),
        azimuthal_angle=jnp.asarray(azimuthal_angle),
        permittivity=jnp.asarray(permittivity_ambient),
    )
    assert in_plane_wavevector.shape[-1] == 2
    assert in_plane_wavevector.ndim == 3

    eigensolve = functools.partial(
        layer.eigensolve_isotropic_media,
        wavelength=jnp.asarray(wavelength),
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
        formulation=fmm.Formulation.FFT,
    )

    mask = unit_cell_pattern(pitch, diameter, resolution)
    permittivity_crystal = jnp.where(mask, permittivity_slab, permittivity_slab)
    solve_result_crystal = eigensolve(permittivity=permittivity_crystal)
    solve_result_ambient = eigensolve(
        permittivity=jnp.asarray(permittivity_ambient)[jnp.newaxis, jnp.newaxis]
    )

    s_matrices_interior = scattering.stack_s_matrices_interior(
        layer_solve_results=[
            solve_result_ambient,
            solve_result_crystal,
            solve_result_ambient,
        ],
        layer_thicknesses=[thickness_ambient_, thickness_slab_, thickness_ambient_],
    )

    # Now compute the eigenmode amplitudes for an incident Gaussian beam.
    # This is done by first obtaining the electric and magnetic fields for the
    # beam, and then solving for the eigenmodes.
    # TODO: replace paraxial Gaussian with a more rigorous expression.

    def _paraxial_gaussian_field_fn(x, y, z):
        # Returns the fields of a z-propagating, x-polarized Gaussian beam.
        # See https://en.wikipedia.org/wiki/Gaussian_beam
        k = 2 * jnp.pi / wavelength
        z_r = jnp.pi * beam_waist**2 * jnp.sqrt(permittivity_ambient) / wavelength
        w_z = beam_waist * jnp.sqrt(1 + (z / z_r) ** 2)
        r = jnp.sqrt(x**2 + y**2)
        ex = (
            beam_waist
            / w_z
            * jnp.exp(-(r**2) / w_z**2)
            * jnp.exp(
                1j
                * (
                    (k * z)  # Phase
                    + k * r**2 / 2 * z / (z**2 + z_r**2)  # Wavefront curvature
                    - jnp.arctan(z / z_r)  # Gouy phase
                )
            )
        )
        ey = jnp.zeros_like(ex)
        ez = jnp.zeros_like(ex)
        hx = jnp.zeros_like(ex)
        hy = ex / jnp.sqrt(permittivity_ambient)
        hz = jnp.zeros_like(ex)
        return (ex, ey, ez), (hx, hy, hz)

    # Solve for the fields of the beam with the desired rotation and shift.
    x, y = basis.unit_cell_coordinates(
        primitive_lattice_vectors=primitive_lattice_vectors,
        shape=permittivity_crystal.shape,  # type: ignore[arg-type]
        num_unit_cells=brillouin_grid_shape,
    )
    (beam_ex, beam_ey, _), (beam_hx, beam_hy, _) = beams.shifted_rotated_fields(
        field_fn=_paraxial_gaussian_field_fn,
        x=x,
        y=y,
        z=jnp.zeros_like(x),
        beam_origin_x=jnp.amax(x) / 2,
        beam_origin_y=jnp.amax(y) / 2,
        beam_origin_z=thickness_ambient_ - beam_focus_offset,
        polar_angle=jnp.asarray(polar_angle),
        azimuthal_angle=jnp.asarray(azimuthal_angle),
        polarization_angle=jnp.asarray(polarization_angle),
    )
    fwd_amplitude, _ = sources.amplitudes_for_fields(
        ex=beam_ex[..., jnp.newaxis],
        ey=beam_ey[..., jnp.newaxis],
        hx=beam_hx[..., jnp.newaxis],
        hy=beam_hy[..., jnp.newaxis],
        layer_solve_result=solve_result_ambient,
        brillouin_grid_axes=(0, 1),
    )

    # Compute the fields inside the structure.
    amplitudes_interior = fields.stack_amplitudes_interior(
        s_matrices_interior=s_matrices_interior,
        forward_amplitude_0_start=fwd_amplitude,
        backward_amplitude_N_end=jnp.zeros_like(fwd_amplitude),
    )
    # Coordinates where fields are to be evaluated.
    x = jnp.arange(0, pitch * brillouin_grid_shape[0], resolution_fields)
    y = jnp.ones_like(x) * pitch * brillouin_grid_shape[1] / 2
    (ex, ey, ez), (hx, hy, hz), (x, y, z) = fields.stack_fields_3d_on_coordinates(
        amplitudes_interior=amplitudes_interior,
        layer_solve_results=[
            solve_result_ambient,
            solve_result_crystal,
            solve_result_ambient,
        ],
        layer_thicknesses=[
            thickness_ambient_,
            thickness_slab_,
            thickness_ambient_,
        ],
        layer_znum=[
            int(thickness_ambient_ / resolution_fields),
            int(thickness_slab_ / resolution_fields),
            int(thickness_ambient_ / resolution_fields),
        ],
        x=x,
        y=y,
    )

    # Perform the Brillouin zone integration by averaging over the Brillouin zone
    # grid batch axes.
    ex, ey, ez, hx, hy, hz = [
        jnp.mean(field, axis=(0, 1)) for field in (ex, ey, ez, hx, hy, hz)
    ]

    # Compute some cross sections for visualizing the structure.
    section_xy, section_xz, section_yz = crystal_cross_sections(
        thickness_ambient=float(thickness_ambient_),
        thickness_slab=float(thickness_slab_),
        pitch=pitch,
        diameter=diameter,
        resolution=resolution,
        num_unit_cells=brillouin_grid_shape,
    )
    return (ex, ey, ez), (hx, hy, hz), (x, y, z), (section_xy, section_xz, section_yz)


def unit_cell_pattern(
    pitch: float,
    diameter: float,
    resolution: float,
) -> jnp.ndarray:
    """Defines the pattern of the photonic crystal."""
    x, y = jnp.meshgrid(
        jnp.arange(0, pitch, resolution),
        jnp.arange(0, pitch, resolution),
        indexing="ij",
    )
    return (jnp.sqrt((x - pitch / 2) ** 2 + y**2) < diameter / 2) | (
        jnp.sqrt((x - pitch / 2) ** 2 + (y - pitch) ** 2) < diameter / 2
    )


def crystal_cross_sections(
    thickness_ambient: float,
    thickness_slab: float,
    pitch: float,
    diameter: float,
    resolution: float,
    num_unit_cells: Tuple[int, int],
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Computes cross sections of the photonic crystal structure."""
    mask = unit_cell_pattern(pitch, diameter, resolution)

    xy_section = jnp.tile(mask, num_unit_cells)

    xz_slab = mask[:, 0]
    xz_section = jnp.stack(
        (
            [jnp.ones_like(xz_slab)] * int(thickness_ambient / resolution)
            + [xz_slab] * int(thickness_slab / resolution)
            + [jnp.ones_like(xz_slab)] * int(thickness_ambient / resolution)
        ),
        axis=-1,
    )
    xz_section = jnp.tile(xz_section, (num_unit_cells[0], 1))

    yz_slab = mask[0, :]
    yz_section = jnp.stack(
        (
            [jnp.ones_like(yz_slab)] * int(thickness_ambient / resolution)
            + [yz_slab] * int(thickness_slab / resolution)
            + [jnp.ones_like(yz_slab)] * int(thickness_ambient / resolution)
        ),
        axis=-1,
    )
    yz_section = jnp.tile(yz_section, (num_unit_cells[1], 1))

    return xy_section, xz_section, yz_section


def plot_dipole_fields(
    pitch: float = PITCH,
    resolution: float = RESOLUTION,
    resolution_fields: float = RESOLUTION_FIELDS,
    brillouin_grid_shape: Tuple[int, int] = BRILLOUIN_GRID_SHAPE,
    **sim_kwargs,
) -> None:
    """Plots an electric field slice for the crystal with embedded source."""
    sim_kwargs.update(
        {
            "pitch": pitch,
            "brillouin_grid_shape": brillouin_grid_shape,
            "resolution": resolution,
            "resolution_fields": resolution_fields,
        }
    )
    (
        (ex, ey, ez),
        (hx, hy, hz),
        (x, y, z),
        (section_xy, section_xz, section_yz),
    ) = simulate_crystal_with_internal_source(**sim_kwargs)

    xplot, zplot = jnp.meshgrid(x, z, indexing="ij")
    field_plot = ex[:, :, 0].real

    plt.figure(figsize=(jnp.amax(xplot), jnp.amax(zplot)), dpi=80)
    ax = plt.subplot(111)
    im = plt.pcolormesh(xplot, zplot, field_plot, shading="nearest", cmap="bwr")

    im.set_clim([-jnp.amax(field_plot), jnp.amax(field_plot)])

    contours = measure.find_contours(onp.array(section_xz))
    scale_factor = pitch / resolution
    for c in contours:
        ax.plot(c[:, 0] / scale_factor, c[:, 1] / scale_factor, "k")

    ax.axis("equal")
    ax.axis("off")
    ax.set_ylim(ax.get_ylim()[::-1])

    plt.subplots_adjust(left=0, bottom=0, right=1, top=1)

    plt.savefig("crystal_dipole.png", bbox_inches="tight")


def plot_gaussian_fields(
    pitch: float = PITCH,
    resolution: float = RESOLUTION,
    resolution_fields: float = RESOLUTION_FIELDS,
    brillouin_grid_shape: Tuple[int, int] = BRILLOUIN_GRID_SHAPE,
    **sim_kwargs,
) -> None:
    """Plots an electric field slice for the crystal with Gaussian beam."""
    sim_kwargs.update(
        {
            "pitch": pitch,
            "brillouin_grid_shape": brillouin_grid_shape,
            "resolution": resolution,
            "resolution_fields": resolution_fields,
        }
    )
    (
        (ex, ey, ez),
        (hx, hy, hz),
        (x, y, z),
        (section_xy, section_xz, section_yz),
    ) = simulate_crystal_with_gaussian_beam(**sim_kwargs)

    xplot, zplot = jnp.meshgrid(x, z, indexing="ij")
    field_plot = ex[:, :, 0].real

    plt.figure(figsize=(jnp.amax(xplot), jnp.amax(zplot)), dpi=80)
    ax = plt.subplot(111)
    im = plt.pcolormesh(xplot, zplot, field_plot, shading="nearest", cmap="bwr")

    im.set_clim([-jnp.amax(field_plot), jnp.amax(field_plot)])

    contours = measure.find_contours(onp.array(section_xz))
    scale_factor = pitch / resolution
    for c in contours:
        ax.plot(c[:, 0] / scale_factor, c[:, 1] / scale_factor, "k")

    ax.axis("equal")
    ax.axis("off")
    ax.set_ylim(ax.get_ylim()[::-1])

    plt.subplots_adjust(left=0, bottom=0, right=1, top=1)

    plt.savefig("crystal_gaussian.png", bbox_inches="tight")


if __name__ == "__main__":
    plot_dipole_fields()
    plot_gaussian_fields()



"""Functions related to Fourier factorization in the FMM algorithm.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import dataclasses
import enum
import functools
from typing import Tuple

import jax
import jax.numpy as jnp

from fmmax import basis, utils, vector


@enum.unique
class Formulation(enum.Enum):
    """Enumerates supported Fourier modal method formulations."""

    FFT: str = "fft"
    JONES_DIRECT: str = vector.JONES_DIRECT
    JONES: str = vector.JONES
    NORMAL: str = vector.NORMAL
    POL: str = vector.POL


# -----------------------------------------------------------------------------
# Functions for computing the Fourier convolution matrices.
# -----------------------------------------------------------------------------


def fourier_matrices_patterned_isotropic_media(
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivity: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: Formulation,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Returns the Fourier convolution matrices for patterned isotropic media.

    All matrices are forms of the Fourier convolution matrices defined in equation
    8 of the S4 reference; as in the reference, we assume the z-axis is separable.
    The matrix for the zz-component is returned directly, while the in-plane
    components are blocked into a single matrix.

    Args:
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivity: The permittivity array, with shape `(..., nx, ny)`.
        expansion: The field expansion to be used.
        formulation: Specifies the formulation to be used.

    Returns:
        eta_matrix: The Fourier convolutio matrix for the inverse of the z-component
            of the permittivity.
        z_permittivity_matrix: The Fourier convolution matrix for the z-component
            of the permittivity.
        transverse_permittivity_matrix: The transverse permittivity matrix from
            equation 15 of [2012 Liu], computed in the manner prescribed by
            `fmm_formulation`.
    """
    if formulation == Formulation.FFT:
        transverse_permittivity_matrix = _transverse_permittivity_fft(
            primitive_lattice_vectors=primitive_lattice_vectors,
            permittivity=permittivity,
            expansion=expansion,
            formulation=formulation,
        )
    else:
        transverse_permittivity_matrix = _transverse_permittivity_vector(
            primitive_lattice_vectors=primitive_lattice_vectors,
            permittivity=permittivity,
            expansion=expansion,
            formulation=formulation,
        )

    transform = functools.partial(
        fourier_convolution_matrix,
        expansion=expansion,
    )

    eta_matrix = transform(1 / permittivity)
    z_permittivity_matrix = transform(permittivity)
    return eta_matrix, z_permittivity_matrix, transverse_permittivity_matrix


def fourier_matrices_patterned_anisotropic_media(
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivity_xx: jnp.ndarray,
    permittivity_xy: jnp.ndarray,
    permittivity_yx: jnp.ndarray,
    permittivity_yy: jnp.ndarray,
    permittivity_zz: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: Formulation,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Returns the Fourier convolution matrices for patterned isotropic media.

    All matrices are forms of the Fourier convolution matrices defined in equation
    8 of the S4 reference; as in the reference, we assume the z-axis is separable.
    The matrix for the zz-component is returned directly, while the in-plane
    components are blocked into a single matrix.

    Args:
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivity_xx: The xx-component of the permittivity tensor, with
            shape `(..., nx, ny)`.
        permittivity_xy: The xy-component of the permittivity tensor.
        permittivity_yx: The yx-component of the permittivity tensor.
        permittivity_yy: The yy-component of the permittivity tensor.
        permittivity_zz: The zz-component of the permittivity tensor.
        expansion: The field expansion to be used.
        formulation: Specifies the formulation to be used.

    Returns:
        eta_matrix: The Fourier convolutio matrix for the inverse of the z-component
            of the permittivity.
        z_permittivity_matrix: The Fourier convolution matrix for the z-component
            of the permittivity.
        transverse_permittivity_matrix: The transverse permittivity matrix from
            equation 15 of [2012 Liu], computed in the manner prescribed by
            `fmm_formulation`.
    """
    del primitive_lattice_vectors
    if formulation != Formulation.FFT:
        raise ValueError(f"Only `Formulation.FFT` is supported, but got {formulation}.")

    transform = functools.partial(
        fourier_convolution_matrix,
        expansion=expansion,
    )

    transverse_permittivity_matrix = jnp.block(
        [
            [transform(permittivity_xx), transform(permittivity_xy)],
            [transform(permittivity_yx), transform(permittivity_yy)],
        ]
    )
    eta_matrix = transform(1 / permittivity_zz)
    z_permittivity_matrix = transform(permittivity_zz)
    return eta_matrix, z_permittivity_matrix, transverse_permittivity_matrix


def _transverse_permittivity_fft(
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivity: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: Formulation,
) -> jnp.ndarray:
    """Computes the `eps` matrix from [2012 Liu] equation 15 using `fft` scheme.

    Args:
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivity: The permittivity array, with shape `(..., nx, ny)`.
        expansion: The field expansion to be used.
        formulation: Specifies the formulation to be used.

    Returns:
        The `eps` matrix.
    """
    del primitive_lattice_vectors
    eps_hat = fourier_convolution_matrix(permittivity, expansion)
    zeros = jnp.zeros_like(eps_hat)
    return jnp.block([[eps_hat, zeros], [zeros, eps_hat]])


def _transverse_permittivity_vector(
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivity: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: Formulation,
) -> jnp.ndarray:
    """Computes transverse permittivity using one of the vector field methods.

    Args:
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivity: The permittivity array, with shape `(..., nx, ny)`.
        expansion: The field expansion to be used.
        formulation: Specifies the formulation to be used.

    Returns:
        The `eps` matrix.
    """
    _transform = functools.partial(fourier_convolution_matrix, expansion=expansion)

    eps_hat = _transform(permittivity)
    zeros = jnp.zeros_like(eps_hat)
    eps_matrix = jnp.block([[eps_hat, zeros], [zeros, eps_hat]])

    eta = 1 / permittivity
    eta_hat = _transform(eta)
    delta_hat = eps_hat - jnp.linalg.inv(eta_hat)
    delta_matrix = jnp.block([[delta_hat, zeros], [zeros, delta_hat]])

    vector_fn = vector.VECTOR_FIELD_SCHEMES[formulation.value]
    tx, ty = vector_fn(permittivity, primitive_lattice_vectors)

    Pxx, Pxy, Pyx, Pyy = tangent_terms(tx, ty)
    p_matrix = jnp.block(
        [[_transform(Pyy), _transform(Pyx)], [_transform(Pxy), _transform(Pxx)]]
    )
    return eps_matrix - delta_matrix @ p_matrix


def tangent_terms(
    tx: jnp.ndarray,
    ty: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Returns the `(Pxx, Pxy, Pyx, Pyy)` from [2012 Liu] equation 50.

    Args:
        tx: The x-component of the tangent vector field.
        ty: The y-component of the tangent vector field.

    Returns:
        The terms `(Pxx, Pxy, Pyx, Pyy)`.
    """
    denom = jnp.abs(tx) ** 2 + jnp.abs(ty) ** 2
    denom_safe = jnp.where(jnp.isclose(denom, 0), 1.0, denom)

    # Compute terms from equation 50 of [2012 Liu]. Note that there is an error
    # in equation 48-51 of [2012 Liu], where x- and y- subscripts are swapped for
    # the diagonal blocks. This can be shown by deriving equation 48 from 46, and
    # can also easily be seen by considering the case of a 1D gratng periodic in
    # the x-direction. In this case, the tangent vector field has a y-component
    # only (i.e. tx = 0), and we want to recover the transverse permittivity
    # given by equation 43. This is possible only by swapping the x- and y-
    # subscripts in the diagonal terms of equations 48-51.
    #
    # Note that we also investigated both orderings numerically with the 1D
    # grating example, and found improved convergence with this correction.
    Pyy = jnp.abs(tx) ** 2 / denom_safe
    Pyx = jnp.conj(tx) * ty / denom_safe
    Pxy = tx * jnp.conj(ty) / denom_safe
    Pxx = jnp.abs(ty) ** 2 / denom_safe
    return Pxx, Pxy, Pyx, Pyy


# -----------------------------------------------------------------------------
# Functions related to Fourier transforms.
# -----------------------------------------------------------------------------


def fourier_convolution_matrix(
    x: jnp.ndarray,
    expansion: basis.Expansion,
) -> jnp.ndarray:
    """Computes the Fourier convolution matrix for `x` and `basis_coefficients`.

    The Fourier convolution matrix at location `(i, j)` gives the Fourier
    coefficient associated with the lattice vector obtained by subtracting the
    `j`th reciprocal lattice vector from the `i`th reciprocal lattice basis.
    See equation 8 from [2012 Liu].

    Args:
        x: The array for which the Fourier coefficients are sought.
        expansion: The field expansion to be used.

    Returns:
        The coefficients, with shape `(num_vectors, num_vectors)`.
    """
    _validate_shape_for_expansion(x.shape, expansion)

    x_fft = jnp.fft.fft2(x)
    x_fft /= jnp.prod(jnp.asarray(x.shape[-2:]))
    idx = _standard_toeplitz_indices(expansion)
    return x_fft[..., idx[..., 0], idx[..., 1]]


def _standard_toeplitz_indices(expansion: basis.Expansion) -> jnp.ndarray:
    """Computes the indices for a standard Toeplitz matrix for `basis_coefficients`.

    Args:
        expansion: The field expansion to be used.

    Returns:
        The indices, with shape `(num, num, 2)`.
    """
    i, j = jnp.meshgrid(
        jnp.arange(expansion.num_terms),
        jnp.arange(expansion.num_terms),
        indexing="ij",
    )
    basis_coefficients = jnp.asarray(expansion.basis_coefficients)
    idx = basis_coefficients[i, :] - basis_coefficients[j, :]
    return idx


def fft(
    x: jnp.ndarray,
    expansion: basis.Expansion,
    axes: Tuple[int, int] = (-2, -1),
) -> jnp.ndarray:
    """Returns the 2D Fourier transform of `x`.

    Args:
        x: The array to be transformed.
        expansion: The field expansion to be used.
        axes: The axes to be transformed, with default being `(-2, -1)`, the final axes.

    Returns:
        The transformed `x`.
    """
    axes: Tuple[int, int] = utils.absolute_axes(axes, x.ndim)  # type: ignore[no-redef]
    _validate_shape_for_expansion(tuple([x.shape[ax] for ax in axes]), expansion)

    x_fft = jnp.fft.fft2(x, axes=axes, norm="forward")

    leading_dims = len(x.shape[: axes[0]])
    trailing_dims = len(x.shape[axes[1] + 1 :])
    slices = (
        [slice(None)] * leading_dims
        + [expansion.basis_coefficients[:, 0], expansion.basis_coefficients[:, 1]]
        + [slice(None)] * trailing_dims
    )
    return x_fft[tuple(slices)]


def ifft(
    y: jnp.ndarray,
    expansion: basis.Expansion,
    shape: Tuple[int, int],
    axis: int = -1,
) -> jnp.ndarray:
    """Returns the 2D inverse Fourier transform of `x`.

    Args:
        y: The array to be transformed.
        expansion: The field expansion to be used.
        shape: The desired shape of the output array.
        axis: The axis containing the Fourier coefficients. Default is `-1`, the
            final axis.

    Returns:
        The inverse transformed `x`.
    """
    (axis,) = utils.absolute_axes((axis,), y.ndim)
    assert y.shape[axis] == expansion.basis_coefficients.shape[-2]
    x_shape = y.shape[:axis] + shape + y.shape[axis + 1 :]
    assert len(x_shape) == y.ndim + 1

    _validate_shape_for_expansion(shape, expansion)

    leading_dims = len(y.shape[:axis])
    trailing_dims = len(y.shape[axis + 1 :])
    slices = (
        [slice(None)] * leading_dims
        + [expansion.basis_coefficients[:, 0], expansion.basis_coefficients[:, 1]]
        + [slice(None)] * trailing_dims
    )

    x = jnp.zeros(x_shape, y.dtype)
    x = x.at[tuple(slices)].set(y)
    return jnp.fft.ifft2(x, axes=(leading_dims, leading_dims + 1), norm="forward")


def _validate_shape_for_expansion(
    shape: Tuple[int, ...],
    expansion: basis.Expansion,
) -> None:
    """Validates that the shape is sufficient for the provided expansion."""
    min_shape = _min_array_shape_for_expansion(expansion)
    if any([d < dmin for d, dmin in zip(shape[-2:], min_shape)]):
        raise ValueError(
            f"`shape` is insufficient for `expansion`, the minimum shape for the "
            f"final two axes is {min_shape} but got shape {shape}."
        )


def _min_array_shape_for_expansion(expansion: basis.Expansion) -> Tuple[int, int]:
    """Returns the minimum allowed shape for an array to be expanded."""
    return (
        int(2 * max(abs(expansion.basis_coefficients[:, 0])) + 1),
        int(2 * max(abs(expansion.basis_coefficients[:, 1])) + 1),
    )


# -----------------------------------------------------------------------------
# Register custom objects in this module with jax to enable `jit`.
# -----------------------------------------------------------------------------


jax.tree_util.register_pytree_node(
    Formulation,
    lambda x: ((), x.value),
    lambda value, _: Formulation(value),
)



"""Tests for `fmmax.sources`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import unittest

import jax
import jax.numpy as jnp
import numpy as onp
import parameterized

from fmmax import basis, fields, fmm, layer, scattering, sources

WAVELENGTH = jnp.array(0.314)
PRIMITIVE_LATTICE_VECTORS = basis.LatticeVectors(u=basis.X, v=basis.Y)
EXPANSION = basis.generate_expansion(
    primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
    approximate_num_terms=20,
    truncation=basis.Truncation.CIRCULAR,
)

LAYER_SOLVE_RESULT = layer.eigensolve_isotropic_media(
    wavelength=WAVELENGTH,
    in_plane_wavevector=jnp.zeros((2,)),
    primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
    permittivity=jnp.ones((50, 50)) * 2,
    expansion=EXPANSION,
    formulation=fmm.Formulation.FFT,
)

BATCH_LAYER_SOLVE_RESULT = layer.eigensolve_isotropic_media(
    wavelength=WAVELENGTH,
    in_plane_wavevector=jnp.zeros((2,)),
    primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
    permittivity=jnp.ones((1, 1, 2, 50, 50)) * 2,
    expansion=EXPANSION,
    formulation=fmm.Formulation.FFT,
)


class FieldSourcesTest(unittest.TestCase):
    def test_amplitudes_match_expected(self):
        # Generate random amplitudes, compute the resulting fields, and extract
        # the amplitudes resulting from those fields. Compare the extracted
        # amplitudes to the original amplitudes.
        brillouin_grid_shape = (5, 5)
        in_plane_wavevector = basis.brillouin_zone_in_plane_wavevector(
            brillouin_grid_shape=brillouin_grid_shape,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
        )
        layer_solve_result = layer.eigensolve_isotropic_media(
            permittivity=jnp.asarray([[1.0]]),
            wavelength=WAVELENGTH,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            expansion=EXPANSION,
            formulation=fmm.Formulation.FFT,
        )
        fwd_amplitude = jax.random.normal(
            jax.random.PRNGKey(0), brillouin_grid_shape + (2 * EXPANSION.num_terms, 3)
        )
        bwd_amplitude = jax.random.normal(
            jax.random.PRNGKey(1), brillouin_grid_shape + (2 * EXPANSION.num_terms, 3)
        )

        efield, hfield = fields.fields_from_wave_amplitudes(
            fwd_amplitude, bwd_amplitude, layer_solve_result
        )
        (ex, ey, _), (hx, hy, _), _ = fields.fields_on_grid(
            efield,
            hfield,
            layer_solve_result,
            shape=(20, 20),
            num_unit_cells=brillouin_grid_shape,
        )
        ex = jnp.mean(ex, axis=(0, 1))
        ey = jnp.mean(ey, axis=(0, 1))
        hx = jnp.mean(hx, axis=(0, 1))
        hy = jnp.mean(hy, axis=(0, 1))
        (
            fwd_amplitude_extracted,
            bwd_amplitude_extracted,
        ) = sources.amplitudes_for_fields(
            ex,
            ey,
            hx,
            hy,
            layer_solve_result,
            brillouin_grid_axes=(0, 1),
        )
        onp.testing.assert_allclose(fwd_amplitude_extracted, fwd_amplitude, rtol=5e-3)
        onp.testing.assert_allclose(bwd_amplitude_extracted, bwd_amplitude, rtol=5e-3)

    def test_field_shape_validation(self):
        brillouin_grid_shape = (5, 5)
        in_plane_wavevector = basis.brillouin_zone_in_plane_wavevector(
            brillouin_grid_shape=brillouin_grid_shape,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
        )
        layer_solve_result = layer.eigensolve_isotropic_media(
            permittivity=jnp.asarray([[1.0]]),
            wavelength=WAVELENGTH,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            expansion=EXPANSION,
            formulation=fmm.Formulation.FFT,
        )
        with self.assertRaisesRegex(
            ValueError, "All fields must be rank 3 with matching shape"
        ):
            sources.amplitudes_for_fields(
                ex=jnp.ones((20, 20, 1)),
                ey=jnp.ones((20, 20, 1)),
                hx=jnp.ones((20, 20, 1)),
                hy=jnp.ones((20, 20)),
                layer_solve_result=layer_solve_result,
                brillouin_grid_axes=(0, 1),
            )

    def test_field_shape_brillouin_grid_compatible_validation(self):
        brillouin_grid_shape = (3, 3)
        in_plane_wavevector = basis.brillouin_zone_in_plane_wavevector(
            brillouin_grid_shape=brillouin_grid_shape,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
        )
        layer_solve_result = layer.eigensolve_isotropic_media(
            permittivity=jnp.asarray([[1.0]]),
            wavelength=WAVELENGTH,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            expansion=EXPANSION,
            formulation=fmm.Formulation.FFT,
        )
        with self.assertRaisesRegex(
            ValueError, "Field shapes must be evenly divisible by the Brillouin"
        ):
            sources.amplitudes_for_fields(
                ex=jnp.ones((20, 20, 1)),
                ey=jnp.ones((20, 20, 1)),
                hx=jnp.ones((20, 20, 1)),
                hy=jnp.ones((20, 20, 1)),
                layer_solve_result=layer_solve_result,
                brillouin_grid_axes=(0, 1),
            )


class InternalSourcesTest(unittest.TestCase):
    @parameterized.parameterized.expand([[(2,)], [(1, 3)], [(1, 3, 2)]])
    def test_gaussian_location_shape_validation(self, invalid_shape):
        with self.assertRaisesRegex(
            ValueError, "`location` must be rank-2 with a trailing axis size of 2"
        ):
            sources.gaussian_source(
                fwhm=0.0,
                location=jnp.ones(invalid_shape),
                in_plane_wavevector=jnp.zeros((1, 2)),
                primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
                expansion=EXPANSION,
            )

    @parameterized.parameterized.expand([[(2,)], [(1, 3)], [(1, 3, 2)]])
    def test_dirac_delta_location_shape_validation(self, invalid_shape):
        with self.assertRaisesRegex(
            ValueError, "`location` must be rank-2 with a trailing axis size of 2"
        ):
            sources.dirac_delta_source(
                location=jnp.ones(invalid_shape),
                in_plane_wavevector=jnp.zeros((1, 2)),
                primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
                expansion=EXPANSION,
            )

    def test_gaussian_wavevector_shape_validation(self):
        with self.assertRaisesRegex(
            ValueError, "`in_plane_wavevector` must have a trailing axis size of 2"
        ):
            sources.gaussian_source(
                fwhm=0.0,
                location=jnp.ones((1, 2)),
                in_plane_wavevector=jnp.zeros((3,)),
                primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
                expansion=EXPANSION,
            )

    def test_dirac_delta_wavevector_shape_validation(self):
        with self.assertRaisesRegex(
            ValueError, "`in_plane_wavevector` must have a trailing axis size of 2"
        ):
            sources.dirac_delta_source(
                location=jnp.ones((1, 2)),
                in_plane_wavevector=jnp.zeros((3,)),
                primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
                expansion=EXPANSION,
            )

    @parameterized.parameterized.expand(
        [
            [(1, 1), 0.4, (0.5, 0.5)],
            [(2, 2), 0.4, (0.5, 0.5)],
            [(3, 4), 0.4, (1.5, 1.5)],
            [(3, 4), 0.5, (1.5, 1.5)],
            [(3, 4), 1.2, (1.5, 1.5)],
        ]
    )
    def test_gaussian_source_matches_expected(
        self, brillouin_grid_shape, fwhm, dipole_location
    ):
        # Performs a simulation of x- and y-oriented Gaussian dipoles in a vacuum,
        # and computes the magnetic field magnitude in the plane of the dipole. We
        # validate that the resulting magnetic field magnitude is actually Gaussian
        # with the specified full width.
        in_plane_wavevector = basis.brillouin_zone_in_plane_wavevector(
            brillouin_grid_shape=brillouin_grid_shape,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
        )
        dipole = sources.gaussian_source(
            fwhm=fwhm,
            location=jnp.asarray([dipole_location]),
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            expansion=EXPANSION,
        )

        zeros = jnp.zeros_like(dipole)
        jx = jnp.concatenate([dipole, zeros], axis=-1)
        jy = jnp.concatenate([zeros, dipole], axis=-1)
        jz = jnp.concatenate([zeros, zeros], axis=-1)

        # Perform the eigendecomposition of a vacuum layer.
        layer_solve_result = layer.eigensolve_isotropic_media(
            permittivity=jnp.asarray([[1.0]]),
            wavelength=WAVELENGTH,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            expansion=EXPANSION,
            formulation=fmm.Formulation.FFT,
        )
        s_matrices_before_source = (
            s_matrices_after_source
        ) = scattering.stack_s_matrices_interior([layer_solve_result], [1.0])
        s_matrix_before_source = s_matrices_before_source[-1][0]
        s_matrix_after_source = s_matrices_after_source[-1][0]
        bwd_amplitude_ambient_end, *_ = sources.amplitudes_for_source(
            jx=jx,
            jy=jy,
            jz=jz,
            s_matrix_before_source=s_matrix_before_source,
            s_matrix_after_source=s_matrix_after_source,
        )

        efield, hfield = fields.fields_from_wave_amplitudes(
            forward_amplitude=jnp.zeros_like(bwd_amplitude_ambient_end),
            backward_amplitude=bwd_amplitude_ambient_end,
            layer_solve_result=layer_solve_result,
        )
        efield, hfield, (x, y) = fields.fields_on_grid(
            electric_field=efield,
            magnetic_field=hfield,
            layer_solve_result=layer_solve_result,
            shape=(21, 21),
            num_unit_cells=brillouin_grid_shape,
        )

        # Perform Brillouin zone integration and compute magnetic field magnitude.
        hfield = jnp.mean(jnp.asarray(hfield), axis=(1, 2))
        hmag = jnp.sqrt(jnp.sum(jnp.abs(hfield) ** 2, axis=0))

        # Add the field for x- and y-oriented dipoles.
        hmag = jnp.sum(hmag, axis=-1)
        hmag /= jnp.amax(hmag)

        self.assertSequenceEqual(hmag.shape, [21 * d for d in brillouin_grid_shape])

        # Compute the expected magnetic field magnitude, which is just a Gaussian.
        x0, y0 = dipole_location
        distance = jnp.sqrt((x - x0) ** 2 + (y - y0) ** 2)
        sigma = fwhm / (2 * jnp.sqrt(2 * jnp.log(2)))
        expected = jnp.exp(-(distance**2) / (2 * sigma**2))

        onp.testing.assert_allclose(hmag, expected, atol=0.05)

    @parameterized.parameterized.expand([[(0, 0)], [(0.5, 0.5)]])
    def test_dirac_delta_matches_gaussian_with_zero_fwhm(self, dipole_location):
        in_plane_wavevector = basis.brillouin_zone_in_plane_wavevector(
            brillouin_grid_shape=(3, 4),
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
        )
        gaussian_dipole = sources.gaussian_source(
            fwhm=0.0,
            location=jnp.asarray([dipole_location]),
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            expansion=EXPANSION,
        )
        dirac_delta_dipole = sources.dirac_delta_source(
            location=jnp.asarray([dipole_location]),
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=PRIMITIVE_LATTICE_VECTORS,
            expansion=EXPANSION,
        )
        onp.testing.assert_allclose(gaussian_dipole, dirac_delta_dipole)


class AmplitudesFromInternalSourcesTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            (1, 0, True, False),
            (0, 1, False, True),
            (1, 1, False, False),
        ]
    )
    def test_emission_planar_source_in_uniform_media(
        self, jx, jy, te_expected_zero, tm_expected_zero
    ):
        # Use the same scattering matrix before and after the source.
        s_matrix = scattering.stack_s_matrix(
            layer_solve_results=(LAYER_SOLVE_RESULT,) * 2,
            layer_thicknesses=(0.2, 0.3),
        )
        (
            bwd_amplitude_0_end,
            fwd_amplitude_before_start,
            bwd_amplitude_before_end,
            fwd_amplitude_after_start,
            bwd_amplitude_after_end,
            fwd_amplitude_N_start,
        ) = sources.amplitudes_for_source(
            jx=fmm.fft(jnp.ones((20, 20, 1)), expansion=EXPANSION, axes=(-3, -2)) * jx,
            jy=fmm.fft(jnp.ones((20, 20, 1)), expansion=EXPANSION, axes=(-3, -2)) * jy,
            jz=fmm.fft(jnp.zeros((20, 20, 1)), expansion=EXPANSION, axes=(-3, -2)),
            s_matrix_before_source=s_matrix,
            s_matrix_after_source=s_matrix,
        )
        with self.subTest("exiting power is nonzero"):
            self.assertFalse(onp.allclose(bwd_amplitude_before_end, 0.0))
            self.assertFalse(onp.allclose(fwd_amplitude_after_start, 0.0))
        with self.subTest("fwd and bwd power at source are equal"):
            onp.testing.assert_allclose(
                jnp.abs(bwd_amplitude_before_end),
                jnp.abs(fwd_amplitude_after_start),
                rtol=1e-05,
                atol=1e-08,
            )
        with self.subTest("bwd power in first layer and at source match"):
            onp.testing.assert_allclose(
                jnp.abs(bwd_amplitude_before_end),
                jnp.abs(bwd_amplitude_0_end),
                rtol=1e-05,
                atol=1e-08,
            )
        with self.subTest("fwd power in last layer and source match"):
            onp.testing.assert_allclose(
                jnp.abs(fwd_amplitude_after_start),
                jnp.abs(fwd_amplitude_N_start),
                rtol=1e-05,
                atol=1e-08,
            )
        with self.subTest("no reflection in the stack"):
            onp.testing.assert_allclose(fwd_amplitude_before_start, 0.0, atol=1e-15)
            onp.testing.assert_allclose(bwd_amplitude_after_end, 0.0, atol=1e-15)
        with self.subTest("check polarization"):
            te, tm = jnp.split(bwd_amplitude_before_end, 2, axis=-2)
            if te_expected_zero:
                onp.testing.assert_allclose(te, 0.0, rtol=1e-05, atol=1e-08)
            else:
                self.assertFalse(onp.allclose(te, 0.0))
            if tm_expected_zero:
                onp.testing.assert_allclose(tm, 0.0, rtol=1e-05, atol=1e-08)
            else:
                self.assertFalse(onp.allclose(tm, 0.0))

    def test_power_xyz_dipole_orientations_match(self):
        def _compute_power(pitch, permittivity, wavelength):
            primitive_lattice_vectors = basis.LatticeVectors(
                u=basis.X * pitch, v=basis.Y * pitch
            )
            layer_solve_result = layer.eigensolve_isotropic_media(
                wavelength=wavelength,
                in_plane_wavevector=jnp.zeros((2,)),
                primitive_lattice_vectors=primitive_lattice_vectors,
                permittivity=jnp.asarray([[permittivity]]),
                expansion=basis.generate_expansion(
                    primitive_lattice_vectors=primitive_lattice_vectors,
                    approximate_num_terms=500,
                    truncation=basis.Truncation.CIRCULAR,
                ),
                formulation=fmm.Formulation.FFT,
            )

            s_matrix = scattering.stack_s_matrix(
                layer_solve_results=(layer_solve_result,),
                layer_thicknesses=(jnp.asarray(pitch / 2),),
            )

            mask = jnp.zeros((100, 100)).at[49:51, 49:51].set(1)

            mask = fmm.fft(mask, expansion=layer_solve_result.expansion)
            zeros = jnp.zeros_like(mask)
            jx = jnp.stack([mask, zeros, zeros], axis=-1)
            jy = jnp.stack([zeros, mask, zeros], axis=-1)
            jz = jnp.stack([zeros, zeros, mask], axis=-1)

            (
                bwd_amplitude_0_end,
                fwd_amplitude_before_start,
                bwd_amplitude_before_end,
                fwd_amplitude_after_start,
                bwd_amplitude_after_end,
                fwd_amplitude_N_start,
            ) = sources.amplitudes_for_source(
                jx=jx,
                jy=jy,
                jz=jz,
                s_matrix_before_source=s_matrix,
                s_matrix_after_source=s_matrix,
            )

            forward_power, _ = fields.amplitude_poynting_flux(
                forward_amplitude=fwd_amplitude_after_start,
                backward_amplitude=jnp.zeros_like(fwd_amplitude_after_start),
                layer_solve_result=layer_solve_result,
            )
            _, backward_power = fields.amplitude_poynting_flux(
                forward_amplitude=jnp.zeros_like(bwd_amplitude_before_end),
                backward_amplitude=bwd_amplitude_before_end,
                layer_solve_result=layer_solve_result,
            )
            forward_power = jnp.sum(forward_power, axis=-2)
            backward_power = jnp.sum(backward_power, axis=-2)
            return forward_power, backward_power

        forward_power, backward_power = _compute_power(
            pitch=2.4, permittivity=1, wavelength=jnp.asarray([0.295, 0.59])
        )

        with self.subTest("Compare x- and y-oriented dipoles"):
            # x- and y-oriented dipoles should match almost exactly.
            onp.testing.assert_allclose(
                forward_power[:, 1], forward_power[:, 0], rtol=1e-6
            )
            onp.testing.assert_allclose(
                backward_power[:, 1], backward_power[:, 0], rtol=1e-6
            )

        with self.subTest("Compare x- and z-oriented dipoles"):
            # x- and z-oriented dipoles may differ, since the structure is periodic in x/y.
            onp.testing.assert_allclose(
                forward_power[:, 2], forward_power[:, 0], rtol=0.2
            )

        # Check that the calculation is scale-invariant as expected. Increase the refractive
        # index by 2, and reduce the pitch by 2. The power should be lower by a factor of 2.
        forward_power_scaled, backward_power_scaled = _compute_power(
            pitch=1.2, permittivity=4, wavelength=jnp.asarray([0.295, 0.59])
        )

        with self.subTest("Check scale invariant"):
            onp.testing.assert_allclose(
                forward_power_scaled, forward_power / 2, rtol=1e-05, atol=1e-08
            )
            onp.testing.assert_allclose(
                backward_power_scaled, backward_power / 2, rtol=1e-05, atol=1e-08
            )

    def test_polarization_terms_shape_validation(self):
        with self.assertRaisesRegex(
            ValueError, "`jx`, `jy`, and `jz` must have matching shape"
        ):
            jx = jnp.ones((20, 20, 1))
            jy = jnp.ones((20, 20, 1))
            jz = jnp.ones((20, 20))
            sources.polarization_terms(jx, jy, jz, LAYER_SOLVE_RESULT)

    @parameterized.parameterized.expand(
        [
            [(20, 30, 1), (-1, 1)],
            [(30, 20, 2), (-1, 2)],
            [(10, 1, 30, 20, 2), (10, 1, -1, 2)],
        ]
    )
    def test_polarization_terms_with_dipole_batch(self, shape, expected_shape):
        pol = sources.polarization_terms(
            jx=fmm.fft(jnp.ones(shape), expansion=EXPANSION, axes=(-3, -2)),
            jy=fmm.fft(jnp.ones(shape), expansion=EXPANSION, axes=(-3, -2)),
            jz=fmm.fft(jnp.ones(shape), expansion=EXPANSION, axes=(-3, -2)),
            layer_solve_result=LAYER_SOLVE_RESULT,
        )
        expected_shape = list(expected_shape)
        expected_shape[-2] = 4 * LAYER_SOLVE_RESULT.expansion.num_terms
        self.assertSequenceEqual(pol.shape, expected_shape)

    @parameterized.parameterized.expand(
        [
            [(20, 30, 1), (1, 1, 2, -1, 1)],
            [(30, 20, 2), (1, 1, 2, -1, 2)],
            [(10, 1, 30, 20, 2), (1, 10, 2, -1, 2)],
        ]
    )
    def test_polarization_terms_with_layer_batch(self, shape, expected_shape):
        pol = sources.polarization_terms(
            jx=fmm.fft(jnp.ones(shape), expansion=EXPANSION, axes=(-3, -2)),
            jy=fmm.fft(jnp.ones(shape), expansion=EXPANSION, axes=(-3, -2)),
            jz=fmm.fft(jnp.ones(shape), expansion=EXPANSION, axes=(-3, -2)),
            layer_solve_result=BATCH_LAYER_SOLVE_RESULT,
        )
        expected_shape = list(expected_shape)
        expected_shape[-2] = 4 * BATCH_LAYER_SOLVE_RESULT.expansion.num_terms
        self.assertSequenceEqual(pol.shape, expected_shape)

    def test_emission_matrix_output_shape(self):
        s_matrix = scattering.stack_s_matrix(
            layer_solve_results=(LAYER_SOLVE_RESULT,) * 2,
            layer_thicknesses=(0.2, 0.3),
        )
        mat = sources.emission_matrix(s_matrix, s_matrix)
        n = LAYER_SOLVE_RESULT.expansion.num_terms
        self.assertSequenceEqual(mat.shape, (4 * n, 4 * n))

    def test_emission_matrix_batch_output_shape(self):
        s_matrix = scattering.stack_s_matrix(
            layer_solve_results=(LAYER_SOLVE_RESULT,) * 2,
            layer_thicknesses=(0.2, 0.3),
        )
        s_matrix_batch = scattering.stack_s_matrix(
            layer_solve_results=(BATCH_LAYER_SOLVE_RESULT,) * 2,
            layer_thicknesses=(0.2, 0.3),
        )
        mat = sources.emission_matrix(s_matrix, s_matrix_batch)
        n = LAYER_SOLVE_RESULT.expansion.num_terms
        self.assertSequenceEqual(mat.shape, (1, 1, 2, 4 * n, 4 * n))



"""A thin film antireflection coating example.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

from typing import Sequence, Tuple

import jax.numpy as jnp
import numpy as onp
import scipy.optimize as spo  # type: ignore[import]

from fmmax import basis, layer, scattering


def compute_reflection(
    refractive_indices: Sequence[jnp.ndarray],
    thicknesses: Sequence[float],
    refractive_index_ambient: jnp.ndarray,
    refractive_index_substrate: jnp.ndarray,
    wavelength: jnp.ndarray,
    incident_angle: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Computes the reflection for a stack of thin films.

    This function can be used to compute batches of reflection coefficients, by
    providing appropriate arrays for the refractive indices, wavelength and
    incident angle.

    Args:
        refractive_indices: The refractive indices for each layer in the stack.
        thicknesses: The thickness for each layer in the stack. Units are arbitrary,
            but should match those of `wavelength`.
        refractive_index_ambient: The refractive index of the ambient, i.e. the
            medium in which the excitation plane wave originates.
        refractive_index_substrate: The refractive index of the substrate.
        wavelength: The wavelengths of the excitation.
        incident_angle: The incident angles of the excitation.

    Returns:
        The reflection for TE- and TM-polarized excitation.
    """

    refractive_indices = [jnp.asarray(r) for r in refractive_indices]
    refractive_index_ambient = jnp.asarray(refractive_index_ambient)
    refractive_index_substrate = jnp.asarray(refractive_index_substrate)

    shape = refractive_index_ambient.shape
    if refractive_index_substrate.shape != shape or any(
        [r.shape != shape for r in refractive_indices]
    ):
        raise ValueError(
            f"All refractive indices must have the same shape, but got "
            f"{[r.shape for r in refractive_indices]}"
            f"{refractive_index_ambient.shape}, and "
            f"{refractive_index_substrate.shape}."
        )

    in_plane_wavevector = basis.plane_wave_in_plane_wavevector(
        wavelength=wavelength,
        polar_angle=incident_angle,
        azimuthal_angle=jnp.zeros_like(incident_angle),
        permittivity=refractive_index_ambient**2,
    )
    primitive_lattice_vectors = basis.LatticeVectors(u=basis.X, v=basis.Y)
    expansion = basis.generate_expansion(
        primitive_lattice_vectors=primitive_lattice_vectors,
        approximate_num_terms=1,
        truncation=basis.Truncation.CIRCULAR,
    )

    thicknesses_with_ambient = (
        [jnp.zeros(1)] + [jnp.asarray(t) for t in thicknesses] + [jnp.zeros(1)]
    )
    permittivities = (
        [jnp.asarray(refractive_index_ambient**2)]
        + [jnp.asarray(n**2) for n in refractive_indices]
        + [jnp.asarray(refractive_index_substrate**2)]
    )
    permittivities = [p[..., jnp.newaxis, jnp.newaxis] for p in permittivities]

    layer_solve_results = [
        layer.eigensolve_uniform_isotropic_media(
            wavelength=wavelength,
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            permittivity=p,
            expansion=expansion,
        )
        for p in permittivities
    ]
    s_matrix = scattering.stack_s_matrix(layer_solve_results, thicknesses_with_ambient)
    r_te = s_matrix.s21[..., 0, 0]
    r_tm = s_matrix.s21[..., expansion.num_terms, expansion.num_terms]
    return r_te, r_tm


def optimize_arc(
    refractive_indices: Tuple[float, ...] = (1.45, 2.0, 1.45, 2.0),
    min_thickness: float = 1.0,
    max_thickness: float = 150.0,
    maxiter: int = 1,
) -> None:
    """Optimizes thicknesses to minimize reflection."""

    def objective(thicknesses: jnp.ndarray) -> float:
        rte, rtm = compute_reflection(
            refractive_indices=[jnp.asarray(r) for r in refractive_indices],
            thicknesses=thicknesses.tolist(),
            refractive_index_ambient=jnp.asarray(1.0),
            refractive_index_substrate=jnp.asarray(1.45),
            wavelength=jnp.asarray([450.0, 520.0, 640.0]),
            incident_angle=jnp.zeros(()),
        )
        return float((jnp.mean(jnp.abs(rte) ** 2) + jnp.mean(jnp.abs(rtm) ** 2)) / 2)

    bounds = onp.asarray([[min_thickness, max_thickness]] * len(refractive_indices))
    result = spo.differential_evolution(
        objective, bounds=bounds, tol=1e-2, maxiter=maxiter
    )

    print(result)

    return result


if __name__ == "__main__":
    optimize_arc()
